import datetime
import os
import sys

import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data import Reader
from mabel.adapters.disk import DiskReader
from rich import traceback
from mabel.data.internals.dictset import STORAGE_CLASS
from mabel.errors import DataNotFoundError

traceback.install()


def test_reader_partitions_read_without_referring_to_partition():
    """
    test if we reference a folder with partitions (by_) without referencing the
    partition, we pick a partition and read it like it's not there
    """
    DATA_DATE = datetime.date(2020, 2, 3)
    records = Reader(
        dataset="tests/data/partitioned",
        inner_reader=DiskReader,
        start_date=DATA_DATE,
        end_date=DATA_DATE,
        persistence=STORAGE_CLASS.MEMORY,
    )
    assert records.count() == 25, records.count()


def test_reader_partitions_read_without_referring_to_partition():
    """
    test if we reference a folder with partitions (by_) without referencing the
    partition, we pick a partition and read it like it's not there
    """
    DATA_DATE = datetime.date(2020, 2, 3)
    records = Reader(
        dataset="tests/data/partitioned",
        inner_reader=DiskReader,
        start_date=DATA_DATE,
        end_date=DATA_DATE,
        persistence=STORAGE_CLASS.MEMORY,
    )
    assert records.count() == 25, records.count()


def test_reader_partitions_read_referring_to_specific_partition():
    """
    test if we reference a folder with partitions (by_) without referencing the
    partition, we pick a partition and read it like it's not there
    """
    DATA_DATE = datetime.date(2020, 2, 3)
    records = Reader(
        dataset="tests/data/partitioned",
        partitions=["year_{yyyy}/month_{mm}/day_{dd}"],
        partition_filter=("userid", "=", "14173315"),
        inner_reader=DiskReader,
        start_date=DATA_DATE,
        end_date=DATA_DATE,
        persistence=STORAGE_CLASS.MEMORY,
    )
    assert records.count() == 5, records.count()

    DATA_DATE = datetime.date(2020, 2, 3)
    records = Reader(
        dataset="tests/data/partitioned",
        partitions=["year_{yyyy}/month_{mm}/day_{dd}"],
        partition_filter=("username", "=", "BBCNews"),
        inner_reader=DiskReader,
        start_date=DATA_DATE,
        end_date=DATA_DATE,
        persistence=STORAGE_CLASS.MEMORY,
    )
    assert records.count() == 4, records.count()

    with pytest.raises(DataNotFoundError):
        DATA_DATE = datetime.date(2020, 2, 3)
        records = Reader(
            dataset="tests/data/partitioned",
            partitions=["year_{yyyy}/month_{mm}/day_{dd}"],
            partition_filter=("username", "=", "CNNNews"),
            inner_reader=DiskReader,
            start_date=DATA_DATE,
            end_date=DATA_DATE,
            persistence=STORAGE_CLASS.MEMORY,
        )
        assert records.count() == 0, records.count()


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import shutil
import datetime
import os
import sys
import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.disk import DiskReader, DiskWriter
from mabel.data import BatchWriter
from mabel.data import Reader
from orso.exceptions import DataValidationError, ExcessColumnsInDataError

dataset = {
    "dataset": "_temp",
    "partitions": ["year_{yyyy}"],
    "format": "parquet",
    "schema": [
        {"name": "character", "type": "VARCHAR", "description": "name"},
        {"name": "persona", "type": "VARCHAR", "description": "named"},
    ],
}


def test_new_schema():
    shutil.rmtree("_temp", ignore_errors=True)
    w = BatchWriter(inner_writer=DiskWriter, date=datetime.datetime.utcnow().date(), **dataset)
    with pytest.raises(ExcessColumnsInDataError) as err:
        w.append(
            {"character": "Barney Stinson", "persona": "Lorenzo Von Matterhorn", "height": 123}
        )
    assert "height" in str(err)

    with pytest.raises(DataValidationError) as err:
        w.append({"persona": "Jackie Daytona"})
    assert "character" in str(err)
    w.finalize()


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import datetime
from typing import Dict
from typing import Optional
from typing import Union

from orso.logging import get_logger

from mabel.data.internals.dictset import STORAGE_CLASS
from mabel.data.internals.dictset import DictSet
from mabel.data.internals.dnf_filters import DnfFilters
from mabel.data.internals.expression import Expression
from mabel.data.readers.internals.cursor import Cursor
from mabel.data.readers.internals.inline_evaluator import Evaluator
from mabel.data.readers.internals.multiprocess_wrapper import processed_reader
from mabel.data.readers.internals.parallel_reader import EXTENSION_TYPE
from mabel.data.readers.internals.parallel_reader import KNOWN_EXTENSIONS
from mabel.data.readers.internals.parallel_reader import ParallelReader
from mabel.data.readers.internals.parallel_reader import pass_thru
from mabel.errors import DataNotFoundError
from mabel.errors import InvalidCombinationError
from mabel.utils.dates import parse_delta
from mabel.utils.parameter_validator import validate

# fmt:off
RULES = [
    {"name": "cursor", "required": False, "warning": None, "incompatible_with": []},
    {"name": "dataset", "required": True, "warning": None, "incompatible_with": []},
    {"name": "end_date", "required": False, "warning": None, "incompatible_with": []},
    {"name": "freshness_limit", "required": False, "warning": None, "incompatible_with": []},
    {"name": "inner_reader", "required": False, "warning": None, "incompatible_with": []},
    {"name": "raw_path", "required": False, "warning": "raw_path will be deprecated, use `partitions` with an empty list", "incompatible_with": ["freshness_limit"]},
    {"name": "select", "required": False, "warning": None, "incompatible_with": []},
    {"name": "start_date", "required": False, "warning": None, "incompatible_with": []},
    {"name": "filters", "required": False, "warning": "", "incompatible_with": []},
    {"name": "persistence", "required": False, "warning": "", "incompatible_with": []},
    {"name": "override_format", "required": False, "warning": "", "incompatible_with": []},
    {"name": "multiprocess", "required": False, "warning": "", "incompatible_with": ["cursor"]},
    {"name": "valid_dataset_prefixes", "required": False},
    {"name": "partitions", "required": False, "warning": None, "incompatible_with": ["raw_path"]},
    {"name": "partition_filter", "required":False, "warning":"`partition_filter` is not expected to be a permanent addition to the API", "incompatible_with": ["freshness_limit"] },
    {"name": "project", "required":False, "warning":"`project` is no longer required for most Readers", "incompatible_with": []}
]
# fmt:on


class AccessDenied(Exception):
    pass


@validate(RULES)
def Reader(
    *,  # force all paramters to be keyworded
    select: str = "*",
    dataset: str = "",
    filters: Optional[str] = None,
    inner_reader=None,  # type:ignore
    raw_path: bool = None,
    persistence: STORAGE_CLASS = STORAGE_CLASS.NO_PERSISTANCE,
    override_format: Optional[str] = None,
    multiprocess: bool = False,
    cursor: Optional[Union[str, Dict]] = None,
    valid_dataset_prefixes: Optional[list] = None,
    partitions=["year_{yyyy}/month_{mm}/day_{dd}"],
    partition_filter=None,
    **kwargs,
) -> DictSet:
    """
    Reads records from a data store, opinionated toward Google Cloud Storage but a
    filesystem reader is available to assist with local development.

    The Reader will iterate over a set of files and return them to the caller as a
    single stream of records. The files can be read from a single folder or can be
    matched over a set of date/time formatted folder names. This is useful to read
    over a set of logs. The date range is provided as part of the call; this is
    essentially a way to partition the data by date/time.

    The reader can filter records to return a subset, for JSON formatted data the
    records can be converted to dictionaries before filtering. JSON data can also
    be used to select columns, so not all read data is returned.

    The reader does not support aggregations, calculations or grouping of data, it
    is a log reader and returns log entries. The reader can convert a set into
    _Pandas_ dataframe, or the _dictset_ helper library can perform some activities
    on the set in a more memory efficient manner.

    Note:
        Different _inner_readers_ may take or require additional parameters. This
        class has a decorator which helps to ensure it is called correctly.

    Parameters:
        select: string (optional):
            A select expression, this is usually a comma separated list of field names
            although can include predefined functions. Default is "*" which presents
            all of the fields in the dataset.
        dataset: string:
            The path to the data source (exact syntax differs per inner_reader)
        filters: string or list/tuple (optional):
            STRING:
            An expression which when evaluated for each row, if False the row will
            be removed from the resulant data set, like the WHERE clause of of a SQL
            statement.
            LIST/TUPLE:
            Filter expressed as DNF.
        inner_reader: BaseReader (optional):
            The reader class to perform the data access Operators, the default is
            GoogleCloudStorageReader
        start_date: datetime (optional):
            The starting date of the range to read over, default is today
        end_date: datetime (optional):
            The end date of the range to read over, default is today
        freshness_limit: string (optional):
            a time delta string (e.g. 6h30m = 6hours and 30 minutes) which
            incidates the maximum age of a dataset before it is no longer
            considered fresh. Where the 'time' of a dataset cannot be
            determined, it will be treated as midnight (00:00) for the date.
        persistence: STORAGE_CLASS (optional)
            How to cache the results, the default is NO_PERSISTANCE which will almost
            always return a generator. MEMORY should only be used where the dataset
            isn't huge and DISK is many times slower than MEMORY. COMPRESSED_MEMORY
            fits in between, usually faster than DISK but slower than MEMORY.
        cursor: dictionary (or string)
            Resume read from a given point (assumes other parameters are the same).
            If a JSON string is provided, it will converted to a dictionary.
        override_format: string (optional)
            Override the format detection - sometimes users know better.
        multiprocess: boolean (optional)
            Split the task over multiple CPUs to improve throughput. Note that there
            are conditions that must be met for the multiprocessor to be safe which
            may mean even though this is set, data is accessed serially.
        valid_dataset_prefixes: list (optional)
            Raises an error if the start of the dataset isn't on the list. The
            intended use is for situations where an external agent can initiate
            the request (such as the Query application). This allows a whitelist
            of allowable resources to be defined.
        partitions: list (optional)
            List of folder names, with datetime placeholders, to use to build a path
            to the data files.
        partition_filter: tuple (optional)
            Provide a hint on how to filter the partitions, as a single tuple in DNF
            notiation, this may be ignored.

    Returns:
        DictSet

    """
    # We can provide an optional whitelist of prefixes that we allow access to
    # - this doesn't replace a proper ACL and permissions model, but can provide
    # some control if other options are limited or unavailable.
    if valid_dataset_prefixes:
        if not any([True for prefix in valid_dataset_prefixes if str(dataset).startswith(prefix)]):
            raise AccessDenied("Access has been denied to this Dataset (prefix).")

    # lazy loading of dependency - in this case the Google GCS Reader
    # eager loading will cause failures when we try to load the google-cloud
    # libraries and they aren't installed.
    if inner_reader is None:
        from ...adapters.google import GoogleCloudStorageReader

        inner_reader = GoogleCloudStorageReader

    # handle transitional states - use the new features to override the legacy features
    if str(raw_path).upper() == "TRUE":
        partitions = None

    # instantiate the injected reader class
    reader_class = inner_reader(
        dataset=dataset,
        partitions=partitions,
        partition_filter=partition_filter,
        **kwargs,
    )  # type:ignore

    arg_dict = kwargs.copy()
    arg_dict["select"] = f"{select}"
    arg_dict["dataset"] = f"{dataset}"
    arg_dict["inner_reader"] = f"{inner_reader.__name__}"  # type:ignore
    arg_dict["filters"] = filters
    get_logger().debug(arg_dict)

    # number of days to walk backwards to find records
    freshness_limit = parse_delta(kwargs.get("freshness_limit", ""))

    if freshness_limit and reader_class.start_date != reader_class.end_date:  # pragma: no cover
        raise InvalidCombinationError(
            "freshness_limit can only be used when the start and end dates are the same"
        )

    return DictSet(
        _LowLevelReader(
            reader_class=reader_class,
            freshness_limit=freshness_limit,
            select=select,
            filters=filters,
            override_format=override_format,
            cursor=cursor,
            multiprocess=multiprocess,
        ),
        storage_class=persistence,
    )


class _LowLevelReader(object):
    def __init__(
        self,
        reader_class,
        freshness_limit,
        select,
        filters,
        override_format,
        cursor,
        multiprocess,
    ):
        self.reader_class = reader_class
        self.freshness_limit = freshness_limit
        self.select = select
        self.override_format = override_format
        self.cursor = cursor
        self._inner_line_reader = None
        self.multiprocess = multiprocess

        if isinstance(filters, str):
            self.filters = Expression(filters)
        elif isinstance(filters, (tuple, list)):
            self.filters = DnfFilters(filters)
        else:
            self.filters = None

        if select != "*":
            self.select = Evaluator(select)
        else:
            self.select = pass_thru

    def _create_line_reader(self):
        # get list of blobs handles as_at, by and frames
        blob_list = self.reader_class.get_list_of_blobs()

        # Handle stepping back if the option is set
        if self.freshness_limit > datetime.timedelta(seconds=1):
            while not bool(blob_list) and self.freshness_limit >= datetime.timedelta(
                days=self.reader_class.days_stepped_back
            ):
                self.reader_class.step_back_a_day()
                blob_list = self.reader_class.get_list_of_blobs()
            if self.freshness_limit < datetime.timedelta(days=self.reader_class.days_stepped_back):
                message = f"No data found in last {self.freshness_limit} - aborting ({self.reader_class.dataset})"
                get_logger().warning(message)
                raise DataNotFoundError(message)
            if self.reader_class.days_stepped_back > 0:
                get_logger().warning(
                    f"Read looked back {self.reader_class.days_stepped_back} day(s) to {self.reader_class.start_date}, limit is {self.freshness_limit} ({self.reader_class.dataset})"
                )

        # Build lists of blobs we have handlers for, based on the file extensions
        supported_blobs = [b for b in blob_list if f".{ b.split('.')[-1] }" in KNOWN_EXTENSIONS]
        readable_blobs = [
            b
            for b in supported_blobs
            if KNOWN_EXTENSIONS[f".{b.split('.')[-1]}"][2] == EXTENSION_TYPE.DATA
        ]

        # Log debug information or an error if there's no blobs to read
        message = f"Reader found {len(readable_blobs)} sources to read data from in `{self.reader_class.dataset}`."
        if len(readable_blobs) == 0:
            get_logger().warning(message)
            raise DataNotFoundError(message)
        else:
            get_logger().debug(message)

        parallel = ParallelReader(
            reader=self.reader_class,
            columns=self.select,
            filters=self.filters or pass_thru,
            override_format=self.override_format,
        )

        use_multiprocess = all(
            [
                self.multiprocess,  # the user must have asked for it
                not self.cursor,  # we must not have a cursor
                len(readable_blobs) > 4,  # we must enough files to read
            ]
        )

        if not use_multiprocess:
            get_logger().debug(f"Serial Reader {self.cursor}")
            if not isinstance(self.cursor, Cursor):
                cursor = Cursor(readable_blobs=readable_blobs, cursor=self.cursor)
                self.cursor = cursor

            blob_to_read = self.cursor.next_blob()
            while blob_to_read:
                blob_reader = parallel(
                    blob_to_read,
                    [
                        idx
                        for idx in supported_blobs
                        if blob_to_read in idx and idx.endswith(".idx")
                    ],
                )
                location = self.cursor.skip_to_cursor(blob_reader)
                for self.cursor.location, record in enumerate(blob_reader, start=location):
                    yield record
                blob_to_read = self.cursor.next_blob(blob_to_read)

        else:
            get_logger().debug("Parallel Reader")
            yield from processed_reader(parallel, readable_blobs, supported_blobs)

    def __iter__(self):
        return self

    def __next__(self):
        if self._inner_line_reader is None:
            self._inner_line_reader = self._create_line_reader()

        # get the the next line from the reader
        return self._inner_line_reader.__next__()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel import Reader
from mabel.data import STORAGE_CLASS
from mabel.adapters.disk import DiskReader
from rich import traceback

traceback.install()


def test_reader_can_read_csv():
    r = Reader(
        inner_reader=DiskReader,
        dataset="tests/data/formats/csv",
        raw_path=True,
        persistence=STORAGE_CLASS.MEMORY,
    )

    # can we read the file into dictionaries
    print(r.first())
    assert r.count() == 33529, r.count()
    assert isinstance(r.first(), dict)

    # are the dictionaries well-formed?
    keys = r.keys(1)
    for row in r:
        assert keys == r.keys(), r.keys()


def test_reader_can_read_zipped_csv():
    r = Reader(
        inner_reader=DiskReader,
        dataset="tests/data/formats/zipped_csv",
        raw_path=True,
        persistence=STORAGE_CLASS.MEMORY,
    )

    # can we read the file into dictionaries
    assert r.count() == 33529, r.count()
    assert isinstance(r.first(), dict)

    # are the dictionaries well-formed?
    keys = r.keys(1)
    for row in r:
        assert keys == r.keys(), r.keys()


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""
Tests for paths to ensure the split and join methods
of paths return the expected values for various
stimulus.
"""
import datetime
import sys
import os
import pathlib

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.utils import paths
from rich import traceback

traceback.install()


def test_blob_paths_get_paths():
    # ignore the filename
    bucket, path, name, ext = paths.get_parts("bucket/parent_folder/sub_folder/filename.ext")
    assert bucket == "bucket"
    assert name == "filename"
    assert ext == ".ext"
    assert path == str(pathlib.PurePosixPath("parent_folder") / "sub_folder") + "/", path

    # with a / at the end
    bucket, path, name, ext = paths.get_parts("bucket/parent_folder/sub_folder/")
    assert bucket == "bucket"
    assert name is None
    assert ext is None
    assert path == str(pathlib.PurePosixPath("parent_folder") / "sub_folder") + "/", path

    # without a / at the end
    bucket, path, name, ext = paths.get_parts("bucket/parent_folder/sub_folder")
    assert bucket == "bucket"
    assert name is None
    assert ext is None
    assert path == str(pathlib.PurePosixPath("parent_folder") / "sub_folder") + "/", path


def test_blob_paths_builder():
    # without trailing /, the / should be added
    template = "year_{yyyy}/month_{mm}/day_{dd}/{yyyy}-{mm}-{dd}/{HH}"
    path = paths.build_path(template, datetime.datetime(2000, 9, 19, 1, 36, 42, 365))
    assert (
        path
        == str(pathlib.PurePosixPath("year_2000") / "month_09" / "day_19" / "2000-09-19" / "01")
        + "/"
    )

    # with trailing /, the / should be retained
    template = "year_{yyyy}/month_{mm}/day_{dd}/{yyyy}-{mm}-{dd}/{HH}/"
    path = paths.build_path(template, datetime.datetime(2000, 9, 19, 1, 36, 42, 365))
    assert (
        path
        == str(pathlib.PurePosixPath("year_2000") / "month_09" / "day_19" / "2000-09-19" / "01")
        + "/"
    )


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""
These are a set of functions that can be applied to data as it passes through.

These are the function definitions, the processor which uses these is in the
'inline_evaluator' module.
"""
import datetime
from functools import lru_cache
from math import trunc

import orjson
from orso.cityhash import CityHash64

from mabel.utils.dates import parse_iso


def get_year(input):
    """
    Convert input to a datetime object and extract the Year part
    """
    if isinstance(input, str):
        input = parse_iso(input)
    if isinstance(input, (datetime.date, datetime.datetime)):
        return input.year
    return None  #


def get_month(input):
    """
    Convert input to a datetime object and extract the Month part
    """
    if isinstance(input, str):
        input = parse_iso(input)
    if isinstance(input, (datetime.date, datetime.datetime)):
        return input.month
    return None


def get_day(input):
    """
    Convert input to a datetime object and extract the Day part
    """
    if isinstance(input, str):
        input = parse_iso(input)
    if isinstance(input, (datetime.date, datetime.datetime)):
        return input.day
    return None


def get_date(input):
    """
    Convert input to a datetime object and extract the Date part
    """
    if isinstance(input, str):
        input = parse_iso(input)
    if isinstance(input, (datetime.date, datetime.datetime)):
        return input.date()
    return None


def get_time(input):
    """
    Convert input to a datetime object and extract the Time part
    """
    if isinstance(input, str):
        input = parse_iso(input)
    if isinstance(input, (datetime.date, datetime.datetime)):
        return input.time()
    return None


def get_quarter(input):
    """
    Convert input to a datetime object and calculate the Quarter of the Year
    """
    if isinstance(input, str):
        input = parse_iso(input)
    if isinstance(input, (datetime.date, datetime.datetime)):
        return ((input.month - 1) // 3) + 1
    return None


def get_hour(input):
    """
    Convert input to a datetime object and extract the Hour part
    """
    if isinstance(input, str):
        input = parse_iso(input)
    if isinstance(input, (datetime.date, datetime.datetime)):
        return input.hour
    return None


def get_minute(input):
    """
    Convert input to a datetime object and extract the Minute part
    """
    if isinstance(input, str):
        input = parse_iso(input)
    if isinstance(input, (datetime.date, datetime.datetime)):
        return input.minute
    return None


def get_second(input):
    """
    Convert input to a datetime object and extract the Seconds part
    """
    if isinstance(input, str):
        input = parse_iso(input)
    if isinstance(input, (datetime.date, datetime.datetime)):
        return input.second
    return None


def get_week(input):
    """
    Convert input to a datetime object and extract the ISO8601 Week
    """
    if isinstance(input, str):
        input = parse_iso(input)
    if isinstance(input, (datetime.date, datetime.datetime)):
        return int(input.strftime("%V"))
    return None


def get_random():
    from ....utils.entropy import random_range

    return random_range(0, 999) / 1000


def concat(*items):
    """
    Turn each item to a string and concatenate the strings together
    """
    sep = ""
    if len(items) == 1 and (isinstance(items[0], (list, tuple, set))):
        items = items[0]
        sep = ", "
    return sep.join(map(str, items))


def to_string(val):
    if isinstance(val, (list, tuple, set)):
        return concat(val)
    if hasattr(val, "mini"):
        return "\\" + val.mini.decode("UTF8") + "\\"
    if isinstance(val, dict):
        return "\\" + orjson.dumps(val).decode("UTF8") + "\\"
    else:
        return str(val)


def add_days(start_date, day_count):
    if isinstance(start_date, str):
        start_date = parse_iso(start_date)
    if isinstance(start_date, (datetime.date, datetime.datetime)):
        return start_date + datetime.timedelta(days=day_count)
    return None


def diff_days(start_date, end_date):
    if isinstance(start_date, str):
        start_date = parse_iso(start_date)
    if isinstance(end_date, str):
        end_date = parse_iso(end_date)
    if isinstance(start_date, (datetime.date, datetime.datetime)) and isinstance(
        end_date, (datetime.date, datetime.datetime)
    ):
        return (end_date - start_date).days
    return None


def parse_number(parser, coerce):
    def inner(val):
        if val is None:
            return None
        return coerce(parser(val))

    return inner


@lru_cache(8)
def get_md5(item):
    import hashlib

    return hashlib.md5(str(item).encode()).hexdigest()  # nosec - meant to be MD5


FUNCTIONS = {
    # DATES & TIMES
    "YEAR": get_year,
    "MONTH": get_month,
    "DAY": get_day,
    "DATE": get_date,
    "QUARTER": get_quarter,
    "WEEK": get_week,
    "HOUR": get_hour,
    "MINUTE": get_minute,
    "SECOND": get_second,
    "TIME": get_time,
    "NOW": datetime.datetime.now,
    "ADDDAYS": add_days,
    "DAYSDIFF": diff_days,
    # STRINGS
    "UCASE": lambda x: str(x).upper(),
    "UPPER": lambda x: str(x).upper(),
    "LCASE": lambda x: str(x).lower(),
    "LOWER": lambda x: str(x).lower(),
    "TRIM": lambda x: str(x).strip(),
    "LEN": len,
    "STRING": to_string,
    "LEFT": lambda x, y: str(x)[: int(y)],
    "RIGHT": lambda x, y: str(x)[-int(y) :],
    "MID": lambda x, y, z: str(x)[int(y) :][: int(z)],
    "CONCAT": concat,
    # NUMBERS
    "ROUND": round,
    "TRUNC": parse_number(float, trunc),
    "INTEGER": parse_number(float, int),
    "DOUBLE": parse_number(float, float),
    # BOOLEAN
    "BOOLEAN": lambda x: str(x).upper() != "FALSE",
    "ISNONE": lambda x: x is None,
    # HASHING & ENCODING
    "HASH": lambda x: format(CityHash64(str(x)), "X"),
    "MD5": get_md5,
    "RANDOM": get_random,  # return a random number 0-99
    # OTHER
    "BETWEEN": lambda val, low, high: low < val < high,
    "SORT": lambda x: sorted(x),
}



from orso import DataFrame
from orso.logging import get_logger


class BaseDatabaseWriter:
    def __init__(self, **kwargs):
        self.logger = get_logger()
        kwargs_passed = [f"{k}={v!r}" for k, v in kwargs.items()]
        self.logger.debug(", ".join(kwargs_passed))

        self.dataset = kwargs.get("dataset")
        self.engine = kwargs.get("engine")
        self.finalizer = kwargs.get("finalizer")
        self.finalized = False

    def commit(self, dataframe: DataFrame):
        raise NotImplementedError("Database")

    def finalize(self):
        # you can either supply your own finalizer via setting the finalizer paramter
        # or you can overide the finalizer in your database writer.
        if self.finalizer and not self.finalized:
            self.finalizer(self)
            self.logger.debug("Database Writer Finalizer")
        self.finalized = True



"""
MinIo Reader - also works with AWS
"""

from ...data.readers.internals.base_inner_reader import BaseInnerReader
from ...errors import MissingDependencyError
from ...utils import dates
from ...utils import paths

try:
    from minio import Minio  # type:ignore

    minio_installed = True
except ImportError:  # pragma: no cover
    minio_installed = False


class MinIoReader(BaseInnerReader):
    RULES = [
        {"name": "end_point", "required": True},
        {"name": "access_key", "required": True},
        {"name": "secret_key", "required": True},
        {"name": "secure", "required": True},
    ]

    def __init__(self, end_point: str, access_key: str, secret_key: str, **kwargs):
        if not minio_installed:  # pragma: no cover
            raise MissingDependencyError(
                "`minio` is missing, please install or include in requirements.txt"
            )

        super().__init__(**kwargs)
        secure = kwargs.get("secure", True)
        self.minio = Minio(end_point, access_key, secret_key, secure=secure)

    def get_blobs_at_path(self, path):
        bucket, object_path, _, _ = paths.get_parts(path)
        for cycle_date in dates.date_range(self.start_date, self.end_date):
            cycle_path = paths.build_path(path=object_path, date=cycle_date)
            blobs = self.minio.list_objects(bucket_name=bucket, prefix=cycle_path, recursive=True)

            yield from [
                bucket + "/" + blob.object_name
                for blob in blobs
                if not blob.object_name.endswith("/")
            ]

    def get_blob_bytes(self, blob_name: str) -> bytes:
        try:
            bucket, object_path, name, extension = paths.get_parts(blob_name)
            stream = self.minio.get_object(bucket, object_path + name + extension)
            return stream.read()
        finally:
            stream.close()



from orso import DataFrame

from .base_database_writer import BaseDatabaseWriter


class NullWriter(BaseDatabaseWriter):
    def commit(self, dataframe: DataFrame):
        # save the data to an interim table
        naive_inserter = f"INSERT INTO `{self.dataset}` ( {', '.join(dataframe.column_names)} )"
        naive_inserter += ",\n".join(f"\t({row})" for row in dataframe) + ";"

        # row acts like a tuple when used
        # it has asdict, asmap (and others), if you want to access the values in the row

        print(naive_inserter)



import time
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.null import NullWriter
from mabel.data import StreamWriter


EXPECTED_RESULTS = {1: 1, 2: 2, 3: 3, 4: 1}


def test_writer_timeout():
    # none of these should do anything
    w = StreamWriter(
        dataset="bucket/path/file.extension",
        inner_writer=NullWriter,
        idle_timeout_seconds=2,
        schema=["data"],
    )

    # with a two second wait, we should timeout and start writing to a new partition
    # after the third write, which has a three second wait (nearly)
    for i in range(1, 5):
        l = w.append({"data": "data"})
        # print(f"cycle {i} - {l} ? {EXPECTED_RESULTS[i]}")
        assert l == EXPECTED_RESULTS[i], f"cycle {i} - {l} != {EXPECTED_RESULTS[i]}"
        if i < len(EXPECTED_RESULTS):
            time.sleep(i * 0.95)


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



# file deepcode ignore ReplaceAPI: test file only
"""
Test the parameter validation on the mabel.data.reader are working
"""
import datetime
import pytest
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data import Reader
from rich import traceback
from mabel.utils import parameter_validator
from mabel.errors import InvalidReaderConfigError
from mabel.data.readers.reader import AccessDenied

traceback.install()


def test_reader_all_good():
    failed = False

    try:
        reader = Reader(
            project="",
            select="a, b",
            dataset="",
            start_date=datetime.datetime.now(),
            end_date=datetime.datetime.now(),
        )
    except InvalidReaderConfigError:
        failed = True

    assert not failed


def test_dataset_prefix_validator():
    with pytest.raises(AccessDenied):
        reader = Reader(dataset="dataset", valid_dataset_prefixes=["drive/"])

    with pytest.raises(AccessDenied):
        reader = Reader(dataset="dataset", valid_dataset_prefixes=["list", "of", "items"])

    # no whitelist - allow all
    reader = Reader(project="", dataset="dataset")

    # a list of one
    reader = Reader(project="", dataset="dataset", valid_dataset_prefixes=["dataset"])

    # a list of many
    reader = Reader(
        project="",
        dataset="dataset",
        valid_dataset_prefixes=["on", "the", "list", "dataset"],
    )


def test_levenshtein():
    ld = parameter_validator.get_levenshtein_distance("abc", "def")
    assert ld == 3, ld

    ld = parameter_validator.get_levenshtein_distance("apples", "pear")
    assert ld == 5, ld

    ld = parameter_validator.get_levenshtein_distance("axe", "ace")
    assert ld == 1, ld


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import os
import sys
import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.disk import DiskReader, DiskWriter
from mabel.adapters.null import NullWriter
from mabel.data import Reader, Writer
from rich import traceback
from data_expectations import Expectations
from data_expectations.errors import ExpectationNotMetError

traceback.install()


VALID_TARGET = {
    "dataset": "_temp",
    "set_of_expectations": [
        {"expectation": "expect_column_to_exist", "column": "name"},
        {"expectation": "expect_column_to_exist", "column": "alter"},
    ],
}

MISSING_COLUMN = {
    "dataset": "_temp",
    "set_of_expectations": [{"expectation": "expect_column_to_exist", "column": "show"}],
}

INVALID_VALUE = {
    "dataset": "_temp",
    "set_of_expectations": [
        {
            "expectation": "expect_column_values_to_be_in_set",
            "column": "name",
            "symbols": ["Laszlo Cravensworth"],
        }
    ],
}


def test_validator_expected_to_work():
    w = Writer(inner_writer=NullWriter, schema=["name", "alter"], **VALID_TARGET)
    w.append({"name": "Barney Stinson", "alter": "Lorenzo Von Matterhorn"})
    w.append({"name": "Laszlo Cravensworth", "alter": "Jackie Daytona"})
    w.append({"name": "Pheobe Buffay", "alter": "Regina Phalange"})
    w.finalize()


def test_validator_expected_to_not_work():
    w = Writer(inner_writer=NullWriter, schema=["name", "alter"], **INVALID_VALUE)
    with pytest.raises(ExpectationNotMetError):
        w.append({"name": "Barney Stinson", "alter": "Lorenzo Von Matterhorn"})
    w.append({"name": "Laszlo Cravensworth", "alter": "Jackie Daytona"})

    w = Writer(inner_writer=NullWriter, schema=["name", "alter"], **MISSING_COLUMN)
    with pytest.raises(ExpectationNotMetError):
        w.append({"name": "Barney Stinson", "alter": "Lorenzo Von Matterhorn"})
    with pytest.raises(ExpectationNotMetError):
        w.append({"name": "Laszlo Cravensworth", "alter": "Jackie Daytona"})
    w.finalize()


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""
Tests for paths to ensure the split and join methods
of paths return the expected values for various
stimulus.
"""
import sys
import os

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.utils import text
from rich import traceback

traceback.install()

QUOTE = """
If you're frightened of dying and you're holding on,
you'll see devils tearing your life away.
But if you've made your peace,
then the devils are really angels,
freeing you from the earth.
-Eckhart von Hochheim
"""

SNIPPET = """
<html>
 <center>
  <blink>
   <b>
    THIS IS IMPORTANT
   </b>
  </blink>
 <center>
</html>
"""


def test_tokenizer():
    assert text.tokenize("the") == ["the"]
    assert text.tokenize("CVE-2017-0144") == ["cve20170144"]
    assert text.tokenize(QUOTE) == [
        "if",
        "youre",
        "frightened",
        "of",
        "dying",
        "and",
        "youre",
        "holding",
        "on",
        "youll",
        "see",
        "devils",
        "tearing",
        "your",
        "life",
        "away",
        "but",
        "if",
        "youve",
        "made",
        "your",
        "peace",
        "then",
        "the",
        "devils",
        "are",
        "really",
        "angels",
        "freeing",
        "you",
        "from",
        "the",
        "earth",
        "eckhart",
        "von",
        "hochheim",
    ]
    assert text.tokenize(SNIPPET) == [
        "html",
        "center",
        "blink",
        "b",
        "this",
        "is",
        "important",
        "b",
        "blink",
        "center",
        "html",
    ]


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""
Test the cursor
"""
import os
import sys
import orjson
import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data.internals.dictset import STORAGE_CLASS
from mabel.adapters.disk import DiskReader
from mabel.adapters.null import NullReader
from mabel.data import Reader
from mabel.utils import entropy
from rich import traceback

traceback.install()


def get_records():
    r = Reader(inner_reader=DiskReader, dataset="tests/data/formats/jsonl", partitions=None)
    return len(list(r))


def test_cursor():
    """
    We're going to test the cursor by doing math on the records, if we don't get the
    right result, it's because we have more or less records than expected.

    To make life easier, we're going to use a literal dataset.
    """

    number_of_records = 10

    data = []
    for i in range(number_of_records):
        data.append({"one": 1, "index": i})

    reader = Reader(inner_reader=NullReader, dataset="none", partitions=None, data=data)

    # create random offsets for testing - it's illogical to have a 0 cursor
    offsets = (entropy.random_range(1, number_of_records) for i in range(20))

    for offset in offsets:
        first_reader = Reader(
            inner_reader=NullReader,
            dataset="none",
            partitions=None,
            data=data,
        )

        # we've going to read to a position in the dataset
        counter = 0
        tracker = []
        for i in range(offset):
            record = next(first_reader)
            counter += record["one"]
            tracker.append(record["index"])

        # we should have read offset records
        assert offset == counter
        # the cursor is zero-based, take one because we've be counting natural numbers
        assert (first_reader.cursor["location"] + 1) == counter

        # we're now going to create a second reader, and give it the cursor from the
        # first
        second_reader = Reader(
            inner_reader=NullReader,
            dataset="none",
            partitions=None,
            data=data,
            cursor=first_reader.cursor,
        )

        # if we keep going, we should get all of the records
        for record in second_reader:
            counter += record["one"]
            tracker.append(record["index"])

        assert counter == number_of_records
        assert tracker == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], tracker


def test_cursor_multiple_times():
    """
    There appears to be a bug with the cursor being used to seek more
    than once in the same file
    """

    number_of_records = 500

    data = []
    for i in range(number_of_records):
        data.append({"one": 1, "index": i})

    cursor = None
    record_counter = 0
    keep_going = True
    times_around = 0

    while keep_going:
        # load the reader
        reader = Reader(
            inner_reader=NullReader,
            dataset="none",
            partitions=None,
            data=data,
            cursor=cursor,
        )

        # read a random number of records
        for burner in range(entropy.random_range(10, 50)):
            try:
                next(reader)
                record_counter += 1
            except:
                keep_going = False

        cursor = reader.cursor
        times_around += 1

    assert record_counter == number_of_records
    assert times_around > 0
    assert cursor.location == -1


def test_cursor_as_text():
    """
    Test that if we pass the cursor as text, it is parsed correctly
    """
    number_of_records = get_records()

    # create random offsets for testing
    offsets = (entropy.random_range(1, number_of_records - 1) for i in range(5))

    for offset in offsets:
        cursor = {"location": offset, "map": "00", "partition": 1983839293359648136}
        reader = Reader(
            inner_reader=DiskReader,
            dataset="tests/data/formats/jsonl",
            partitions=None,
            cursor=orjson.dumps(cursor).decode(),
        )

        records_left = len(list(reader))
        # the offset is zero-based, for example it says 10, but it's 11 records
        assert (
            offset + 1
        ) + records_left == number_of_records, f"{offset + 1 + records_left} != {number_of_records}"


def test_move_to_cursor():
    """
    Test when we move to a cursor position, that we read all of the records
    """
    number_of_records = get_records()

    # create random offsets for testing
    offsets = (entropy.random_range(1, number_of_records - 1) for i in range(5))

    for offset in offsets:
        reader = Reader(
            inner_reader=DiskReader,
            dataset="tests/data/formats/jsonl",
            partitions=None,
            cursor={"location": offset, "map": "00", "partition": 1983839293359648136},
            persistence=STORAGE_CLASS.NO_PERSISTANCE,
        )

        assert offset == reader.cursor["location"], f"{offset}, {reader.cursor['location']}"

        records_left = len(list(reader))
        # the offset is zero-based, for example it says 10, but it's 11 records
        assert (
            offset + 1
        ) + records_left == number_of_records, f"{offset + 1 + records_left} != {number_of_records}"


def test_zero_based_cursor():
    """
    Test that the cursor is actually zero-based
    """
    reader = Reader(inner_reader=DiskReader, dataset="tests/data/formats/jsonl", partitions=[])
    number_of_records = get_records()

    # we read all of the records, one-by-one
    for i in range(number_of_records):
        next(reader)
        assert reader.cursor["location"] == i, reader.cursor

    # we can't read past the end
    with pytest.raises(StopIteration):
        next(reader)

    # test our assumptions about the 'range' function
    assert len(range(number_of_records)) == number_of_records
    assert list(range(5)) == [0, 1, 2, 3, 4]


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import sys
import os

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data.internals.group_by import GroupBy
from mabel.data.internals.dictset import STORAGE_CLASS, DictSet
from rich import traceback
from decimal import Decimal

traceback.install()


def test_group_by():
    """
    Test simple grouping and aggregation
    """
    data = [
        {"user": "bob", "value": 1},
        {"user": "bob", "value": 2},
        {"user": "bob", "value": 1},
        {"user": "bob", "value": 2},
        {"user": "bob", "value": 1},
        {"user": "alice", "value": 3},
        {"user": "alice", "value": 4},
        {"user": "alice", "value": 3},
        {"user": "alice", "value": 4},
        {"user": "alice", "value": 5},
        {"user": "alice", "value": 5},
        {"user": "eve", "value": 6},
        {"user": "eve", "value": 7},
    ]

    # fmt:off
    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    assert list(GroupBy(ds, "user").groups()) == [{'user': 'bob'}, {'user': 'alice'}, {'user': 'eve'}]

    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    gb = GroupBy(ds, "user")
    cn = gb.count()
    ls = list(cn)
    assert ls == [{'COUNT(*)': 6, 'user': 'alice'}, {'COUNT(*)': 5, 'user': 'bob'}, {'COUNT(*)': 2, 'user': 'eve'}], ls

    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    gb = list(GroupBy(ds, "user").average("value"))
    assert gb == [{'AVG(value)': Decimal("4.0"), 'user': 'alice'}, {'AVG(value)': Decimal("1.4"), 'user': 'bob'}, {'AVG(value)': Decimal("6.5"), 'user': 'eve'}], gb

    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    assert list(GroupBy(ds, "user").max("value")) == [{'MAX(value)': 5, 'user': 'alice'}, {'MAX(value)': 2, 'user': 'bob'}, {'MAX(value)': 7, 'user': 'eve'}]

    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    assert list(GroupBy(ds, "user").min("value")) == [{'MIN(value)': 3, 'user': 'alice'}, {'MIN(value)': 1, 'user': 'bob'}, {'MIN(value)': 6, 'user': 'eve'}]
    # fmt:on


def test_combined_group_by():
    """
    Test combined grouping and aggregation
    """
    data = [
        {"fname": "bob", "sname": "smith", "value": 1, "cost": 1},
        {"fname": "bob", "sname": "jones", "value": 2, "cost": 4},
        {"fname": "bob", "sname": "smith", "value": 1, "cost": 3},
        {"fname": "bob", "sname": "jones", "value": 2, "cost": 2},
        {"fname": "bob", "sname": "smith", "value": 1, "cost": 1},
        {"fname": "alice", "sname": "jones", "value": 3, "cost": 4},
        {"fname": "alice", "sname": "smith", "value": 4, "cost": 3},
        {"fname": "alice", "sname": "jones", "value": 3, "cost": 2},
        {"fname": "alice", "sname": "smith", "value": 4, "cost": 1},
        {"fname": "alice", "sname": "jones", "value": 5, "cost": 4},
        {"fname": "alice", "sname": "smith", "value": 5, "cost": 3},
        {"fname": "eve", "sname": "jones", "value": 6, "cost": 2},
        {"fname": "eve", "sname": "smith", "value": 7, "cost": 1},
    ]

    # fmt:off
    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    gs = list(GroupBy(ds, ("fname", "sname")).groups())
    assert gs == [{'fname': 'bob', 'sname': 'smith'}, {'fname': 'bob', 'sname': 'jones'}, {'fname': 'alice', 'sname': 'jones'}, {'fname': 'alice', 'sname': 'smith'}, {'fname': 'eve', 'sname': 'jones'}, {'fname': 'eve', 'sname': 'smith'}], gs

    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    gs = list(GroupBy(ds, ("fname", "sname")).count())
    assert gs ==  [{'COUNT(*)': 2, 'fname': 'bob', 'sname': 'jones'}, {'COUNT(*)': 3, 'fname': 'bob', 'sname': 'smith'}, {'COUNT(*)': 3, 'fname': 'alice', 'sname': 'smith'}, {'COUNT(*)': 1, 'fname': 'eve', 'sname': 'smith'}, {'COUNT(*)': 1, 'fname': 'eve', 'sname': 'jones'}, {'COUNT(*)': 3, 'fname': 'alice', 'sname': 'jones'}], gs

    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    gs = list(GroupBy(ds, ("fname", "sname")).average('value'))
    assert gs == [{'AVG(value)': Decimal('2.0'), 'fname': 'bob', 'sname': 'jones'}, {'AVG(value)': Decimal('1.0'), 'fname': 'bob', 'sname': 'smith'}, {'AVG(value)': Decimal('4.333333333333333333333333333'), 'fname': 'alice', 'sname': 'smith'}, {'AVG(value)': Decimal('7.0'), 'fname': 'eve', 'sname': 'smith'}, {'AVG(value)': Decimal('6.0'), 'fname': 'eve', 'sname': 'jones'}, {'AVG(value)': Decimal('3.666666666666666666666666667'), 'fname': 'alice', 'sname': 'jones'}], gs

    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    gs = list(GroupBy(ds, ("fname", "sname")).average('cost'))
    assert gs == [{'AVG(cost)': Decimal('3.0'), 'fname': 'bob', 'sname': 'jones'}, {'AVG(cost)': Decimal('1.666666666666666666666666667'), 'fname': 'bob', 'sname': 'smith'}, {'AVG(cost)': Decimal('2.333333333333333333333333333'), 'fname': 'alice', 'sname': 'smith'}, {'AVG(cost)': Decimal('1.0'), 'fname': 'eve', 'sname': 'smith'}, {'AVG(cost)': Decimal('2.0'), 'fname': 'eve', 'sname': 'jones'}, {'AVG(cost)': Decimal('3.333333333333333333333333333'), 'fname': 'alice', 'sname': 'jones'}], gs

    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    gs = list(GroupBy(ds, ("fname", "sname")).average(('cost', 'value',)))
    assert gs == [{'AVG(cost)': Decimal('3'), 'AVG(value)': Decimal('2'), 'fname': 'bob', 'sname': 'jones'}, {'AVG(cost)': Decimal('1.666666666666666666666666667'), 'AVG(value)': Decimal('1'), 'fname': 'bob', 'sname': 'smith'}, {'AVG(cost)': Decimal('2.333333333333333333333333333'), 'AVG(value)': Decimal('4.333333333333333333333333333'), 'fname': 'alice', 'sname': 'smith'}, {'AVG(cost)': 1.0, 'AVG(value)': 7.0, 'fname': 'eve', 'sname': 'smith'}, {'AVG(cost)': 2.0, 'AVG(value)': 6.0, 'fname': 'eve', 'sname': 'jones'}, {'AVG(cost)': Decimal('3.333333333333333333333333333'), 'AVG(value)': Decimal('3.666666666666666666666666667'), 'fname': 'alice', 'sname': 'jones'}]

    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    gs = list(GroupBy(ds, ("fname", "sname")).aggregate([('MAX', 'value'),('MIN', 'cost')]))
    assert gs == [{'MAX(value)': 2, 'MIN(cost)': 2, 'fname': 'bob', 'sname': 'jones'}, {'MAX(value)': 1, 'MIN(cost)': 1, 'fname': 'bob', 'sname': 'smith'}, {'MAX(value)': 5, 'MIN(cost)': 1, 'fname': 'alice', 'sname': 'smith'}, {'MAX(value)': 7, 'MIN(cost)': 1, 'fname': 'eve', 'sname': 'smith'}, {'MAX(value)': 6, 'MIN(cost)': 2, 'fname': 'eve', 'sname': 'jones'}, {'MAX(value)': 5, 'MIN(cost)': 2, 'fname': 'alice', 'sname': 'jones'}], gs

    # fmt:on


def test_gappy_set():
    data = [
        {"key": 1, "value": "one", "plus1": 2},
        {"key": 2, "value": "two", "plus1": 3},
        {"key": None, "value": "one", "plus1": 4},
        {"key": 4, "value": "two", "plus1": 5},
        {"key": 4, "value": None, "plus1": 5},
    ]
    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    g = list(ds.group_by("value").average("key"))
    assert g == [
        {"AVG(key)": 4.0, "value": None},
        {"AVG(key)": 3.0, "value": "two"},
        {"AVG(key)": 1.0, "value": "one"},
    ], g


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



from .internals.dictset import STORAGE_CLASS
from .internals.dictset import DictSet
from .readers.internals.sql_reader import SqlReader
from .readers.reader import Reader
from .writers.batch_writer import BatchWriter
from .writers.database_writer import DatabaseWriter
from .writers.stream_writer import StreamWriter
from .writers.writer import Writer



# no-maintain-checks
"""
There are multiple usecases where we need to step over a set of tokens and apply
a label to them. Doing this in a helper module means a) it only needs to be maintained
in one place b) different parts of the system behave consistently.
"""

import operator
import re
from functools import lru_cache

from mabel.data.internals.group_by import AGGREGATORS
from mabel.data.readers.internals.inline_functions import FUNCTIONS
from mabel.utils.dates import parse_iso
from mabel.utils.text import like
from mabel.utils.text import matches
from mabel.utils.text import not_like

# These are the characters we should escape in our regex
REGEX_CHARACTERS = {ch: "\\" + ch for ch in ".^$*+?{}[]|()\\"}


class TokenError(Exception):
    pass


def interpret_value(value):
    if not isinstance(value, str):
        return value
    if value.upper() in ("TRUE", "FALSE"):
        return value.upper() == "TRUE"

    try:
        num = int(value)
        return num
    except ValueError:
        pass

    try:
        num = float(value)
        return num
    except ValueError:
        pass

    value = value[1:-1]
    return parse_iso(value) or value


def function_in(x, y):
    candidates = [interpret_value(i) for i in y if str(i).strip() != ","]

    return x in candidates


def function_contains(x, y):
    return y in x


# the order of the operators affects the regex, e.g. <> needs to be defined before
# < otherwise that will be matched and the > will be invalid syntax.
OPERATORS = {
    "<>": operator.ne,
    ">=": operator.ge,
    "<=": operator.le,
    ">": operator.gt,
    "<": operator.lt,
    "==": operator.eq,
    "!=": operator.ne,
    "=": operator.eq,
    "IS NOT": operator.is_not,
    "IS": operator.is_,
    "NOT LIKE": not_like,
    "LIKE": like,
    "MATCHES": matches,
    "IN": function_in,
    "CONTAINS": function_contains,
}


class TOKENS(int):
    UNKNOWN = -1
    INTEGER = 0
    FLOAT = 1
    LITERAL = 2
    VARIABLE = 3
    BOOLEAN = 4
    DATE = 5
    NULL = 6
    LEFTPARENTHESES = 7
    RIGHTPARENTHESES = 8
    COMMA = 9
    FUNCTION = 10
    AGGREGATOR = 11
    AS = 12
    EVERYTHING = 13
    OPERATOR = 14
    AND = 15
    OR = 16
    NOT = 17
    SUBQUERY = 18
    EMPTY = 19
    LIST = 20


@lru_cache(64)
def get_token_type(token):
    """
    Guess the token type.
    """
    token = str(token).strip()
    token_upper = token.upper()
    if len(token) == 0:
        return TOKENS.EMPTY
    if token[0] == token[-1] == "`":
        # tokens in ` quotes are variables, this is how we supersede all other
        # checks, e.g. if it looks like a number but is a variable.
        return TOKENS.VARIABLE
    if token == "*":  # nosec - not a password
        return TOKENS.EVERYTHING
    if token_upper in FUNCTIONS:
        return TOKENS.FUNCTION
    if token_upper in OPERATORS:
        return TOKENS.OPERATOR
    if token_upper in AGGREGATORS:
        return TOKENS.AGGREGATOR
    if token[0] == token[-1] == '"' or token[0] == token[-1] == "'":
        # tokens in quotes are either dates or string literals, if we can
        # parse to a date, it's a date
        if parse_iso(token[1:-1]):
            return TOKENS.DATE
        else:
            return TOKENS.LITERAL

    try:
        num = int(token)
        return TOKENS.INTEGER
    except ValueError:
        pass

    try:
        num = float(token)
        return TOKENS.FLOAT
    except ValueError:
        pass

    if token in ("(", "["):
        return TOKENS.LEFTPARENTHESES
    if token in (")", "]"):
        return TOKENS.RIGHTPARENTHESES
    if token == ",":  # nosec
        return TOKENS.COMMA
    if re.search(r"^[^\d\W][\w\-\.]*", token):
        if token_upper in ("TRUE", "FALSE"):
            # 'true' and 'false' without quotes are booleans
            return TOKENS.BOOLEAN
        if token_upper in ("NULL", "NONE"):
            # 'null' or 'none' without quotes are nulls
            return TOKENS.NULL
        if token_upper == "AND":  # nosec - not a password
            return TOKENS.AND
        if token_upper == "OR":  # nosec - not a password
            return TOKENS.OR
        if token_upper == "NOT":  # nosec - not a password
            return TOKENS.NOT
        if token_upper == "AS":  # nosec - not a password
            return TOKENS.AS
        # tokens starting with a letter, is made up of letters, numbers,
        # hyphens, underscores and dots are probably variables. We do this
        # last so we don't miss assign other items to be a variable
        return TOKENS.VARIABLE
    # at this point, we don't know what it is
    return TOKENS.UNKNOWN


@lru_cache(1)
def build_splitter():
    # build the regex by building a list of all of the keywords
    keywords = []
    for item in FUNCTIONS:
        keywords.append(r"\b" + item + r"\b")
    for item in AGGREGATORS:
        keywords.append(r"\b" + item + r"\b")
    for item in OPERATORS:
        if item.replace(" ", "").isalpha():
            keywords.append(r"\b" + item + r"\b")
        else:
            keywords.append("".join([REGEX_CHARACTERS.get(ch, ch) for ch in item]))
    for item in [
        "AND",
        "OR",
        "NOT",
        "SELECT",
        "FROM",
        "WHERE",
        "LIMIT",
        r"GROUP\sBY",
        r"ORDER\sBY",
        "DISTINCT",
        "ASC",
        "DESC",
        "IN",
    ]:
        keywords.append(r"\b" + item + r"\b")
    for item in ["(", ")", "[", "]", ",", "*"]:
        keywords.append("".join([REGEX_CHARACTERS.get(ch, ch) for ch in item]))
    splitter = re.compile(
        r"(" + r"|".join(keywords) + r")",
        re.IGNORECASE,
    )
    return splitter


class Tokenizer:
    slots = ("i", "tokens")

    def __init__(self, exp):
        self.i = 0
        if isinstance(exp, str):
            self.tokens = self.tokenize(exp)
        else:
            self.tokens = exp

    def next(self):
        self.i += 1
        return self.tokens[self.i - 1]

    def peek(self):
        return self.tokens[self.i]

    def has_next(self):
        return self.i < len(self.tokens)

    def next_token_type(self):
        return get_token_type(self.tokens[self.i])

    def next_token_value(self):
        return self.tokens[self.i]

    def _fix_special_chars(self, tokens):
        """
        The splitter will cut quoted tokens if they contain characters or strings
        we generally want to split on - quotes can be used to say, this isn't an
        instance of that character to split on, so we join these quoted strings
        back together.
        """

        builder = ""
        looking_for_end_char = None

        for token in tokens:
            stripped_token = token.strip()
            if len(stripped_token) == 0:
                # the splitter creates unwanted empty strings
                pass

            elif not looking_for_end_char and stripped_token[0] not in ('"', "'", "`"):
                # nothing interesting here
                yield token

            elif (
                stripped_token[0] in ('"', "'", "`")
                and stripped_token[-1] == stripped_token[0]
                and len(stripped_token) > 1
            ):
                # the quotes wrap the entire token
                yield token

            elif stripped_token[-1] == looking_for_end_char:
                # we've found the end of the token, yield it and reset
                builder += token
                yield builder
                builder = ""
                looking_for_end_char = None

            elif stripped_token[0] in ('"', "'", "`") and (
                stripped_token[-1] != stripped_token[0] or len(stripped_token) == 1
            ):
                # we've found a new token to collect
                # the last character will always equal the last character if there's only one
                builder = token
                looking_for_end_char = stripped_token[0]

            elif looking_for_end_char:
                # we're building a token
                builder += token

            else:
                raise TokenError(
                    "Unable to determine quoted token boundaries, you may be missing a closing quote."
                )

    def _case_correction(self, tokens):
        for token in tokens:
            if get_token_type(token) in (
                TOKENS.LITERAL,
                TOKENS.VARIABLE,
                TOKENS.SUBQUERY,
            ):
                yield token
            else:
                yield token.upper()

    def clean_statement(self, string):
        """
        Remove carriage returns and all whitespace to single spaces
        """
        whitespace_cleaner = re.compile(r"\s+")
        return whitespace_cleaner.sub(" ", string).strip()

    def tokenize(self, expression):
        expression = self.clean_statement(expression)
        tokens = build_splitter().split(expression)
        # characters like '*' in literals break the tokenizer, so we need to fix them
        tokens = list(self._fix_special_chars(tokens))
        tokens = [t.strip() for t in tokens if t.strip() != ""]
        tokens = list(self._case_correction(tokens))
        return tokens

    def __str__(self):
        return self.peek()



import pytest
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel import Reader
from mabel.data import STORAGE_CLASS
from mabel.adapters.disk import DiskReader
from rich import traceback

traceback.install()


from orso.logging import get_logger

get_logger().setLevel(5)


def test_reader_can_read():
    r = Reader(inner_reader=DiskReader, dataset="tests/data/tweets", raw_path=True)
    assert len(list(r)) == 50


def test_reader_to_pandas():
    r = Reader(inner_reader=DiskReader, dataset="tests/data/tweets", raw_path=True)
    df = r.to_pandas()

    assert len(df) == 50


def test_reader_can_read_alot():
    r = Reader(inner_reader=DiskReader, dataset="tests/data/nvd", raw_path=True)
    for i, row in enumerate(r):
        if i % 1000000 == 0:
            print(i)
        pass
    assert i == 135133 or i == 16066287, i


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



from typing import Callable

from mabel.data.internals.dictset import STORAGE_CLASS
from mabel.data.internals.dictset import DictSet


class CollectedSet:
    __slots__ = ["_collections"]

    def __init__(self, dictset: DictSet, column: str, dedupe: bool = False):
        """
        Collection functionality for Iterables of Dictionaries
        Parameters:
            dictset: Iterable of dictionaries
                The dataset to perform the Collection on
            column: string
                The name of the field to collect by
            dedupe: bool (optional)
                Remove duplicate values from the collections
        Returns:
            CollectedSet
        Warning:
            The 'Collection' object holds the entire dataset in memory so is
            unsuitable for large datasets.
        """
        collections: dict = {}

        groups = dictset.distinct() if dedupe else dictset

        for item in groups:
            my_item = item.copy()
            key = my_item.pop(column, None)
            collections.setdefault(key, []).append(my_item)
        if dedupe:
            collections = {
                k: {frozenset(i.items()): i for i in v}.values() for k, v in collections.items()
            }

        self._collections = collections

    def count(self, collection=None):
        """
        Count the number of items in collections
        Parameters:
            collection: string (optional)
                If provided, return the count of just this collection
        Returns:
            if a collection is provided, return an integer
            if no collection is provided, return a dictionary
        """
        if collection is None:
            return {x: len(y) for x, y in self._collections.items()}
        else:
            try:
                return [len(y) for x, y in self._collections.items() if x == collection].pop()
            except IndexError:
                return 0

    def aggregate(self, column, method):
        """
        Applies an aggregation function by collection.
        Parameters:
            column: string
                The name of the field to aggregate on
            method: callable
                The function to aggregate with
        Returns:
            dictionary
        Examples:
            maxes = collection.aggregate('age', max)
            means = collection.aggregate('age', maths.mean)
        """
        response = {}
        for key, items in self._collections.items():
            values = [item.get(column) for item in items if item.get(column) is not None]
            response[key] = method(values)
        return response

    def apply(self, method: Callable):
        """
        Apply a function to all collections
        Parameters:
            method: callable
                The function to apply to the collections
        Returns:
            dictionary
        """
        return {key: method(items) for key, items in self._collections.items()}

    def __len__(self):
        """
        Returns the number of groups in the set.
        """
        return len(self._collections)

    def __repr__(self):
        """
        Returns the group names
        """
        return f"Collection of {len(self)} items"

    def __contains__(self, item):
        return item in self._collections

    def __getitem__(self, item):
        """
        Selector access to groups, e.g. Groups["Group Name"]
        Note that Groups["Group 1", "Group 2"] creates a group with just those items
        """
        if isinstance(item, (tuple, list)):
            newg = CollectedSet([], None)
            for entry in item:
                if entry in self._collections:
                    newg._collections[entry].append(self._collections[entry])
            return newg
        else:
            return SubCollection(self._collections.get(item))

    def items(self):
        for collection in self._collections:
            for item in self._collections[collection]:
                yield collection, item


class SubCollection:
    __slots__ = ["values"]

    def __init__(self, values):
        self.values = DictSet(values or [], storage_class=STORAGE_CLASS.MEMORY)

    def __getitem__(self, item):
        """
        Selector access to a value in a collection, support arrays
        """
        if isinstance(item, tuple):
            return list(self.values.select(*item))
        else:
            return self.values.collect_list(item)

    def __len__(self):
        return self.values.count()

    def __repr__(self):
        return f"SubCollection of {len(self)} items"

    def get(self, item):
        values = self[item]
        if len(values) == 0 or values is None:
            return None
        return values



import datetime
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel import Reader
from mabel.adapters.disk import DiskReader
from rich import traceback

traceback.install()


START_DATE = datetime.datetime(2020, 12, 21, 18)
END_DATE = datetime.datetime(2020, 12, 22, 1)


def test_reader_times():
    subject = Reader(
        dataset="tests/data/dated",
        partitions=["year_{yyyy}/month_{mm}/day_{dd}/by_hour/hour={HH}"],
        start_date=START_DATE,
        end_date=END_DATE,
        inner_reader=DiskReader,
    )

    records = subject.collect_list()
    assert len(records) == 4, len(records)


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import os
import sys

import orjson

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.null import NullWriter
from mabel.data import StreamWriter
from rich import traceback

traceback.install()


DATA_SET = [
    {"key": 6, "value": ["s", "i", "x"], "combinations": 3},  # (6,s),(6,i),(6,x)
    {"key": [1, 0], "value": ["t", "e", "n"], "combinations": 6},
    {"key": 0, "combinations": 1},
]


def test_writer_substitutions():
    w = StreamWriter(dataset="TEST/{key}/{value}", inner_writer=NullWriter, schema=["@"])

    for record in DATA_SET:
        # convert to a simd object to test behavior
        as_json = orjson.dumps(record)

        combinations = w.append(record)
        assert combinations == record["combinations"], combinations


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import datetime
import re
from typing import Optional
from typing import Union

try:
    import numpy
except ImportError:
    numpy = None

TIMEDELTA_REGEX = (
    r"((?P<days>-?\d+)d)?"
    r"((?P<hours>-?\d+)h)?"
    r"((?P<minutes>-?\d+)m)?"
    r"((?P<seconds>-?\d+)s)?"
)
TIMEDELTA_PATTERN = re.compile(TIMEDELTA_REGEX, re.IGNORECASE)
SECONDS_PER_HOUR: int = 3600


# based on:
# https://gist.github.com/santiagobasulto/698f0ff660968200f873a2f9d1c4113c#file-parse_timedeltas-py
def parse_delta(delta: str) -> datetime.timedelta:
    """
    Parses a human readable timedelta (3d5h19m) into a datetime.timedelta.

    Delta includes:
    * Xd days
    * Xh hours
    * Xm minutes
    * Xs seconds

    Values can be negative following timedelta's rules. Eg: -5h-30m
    """
    match = TIMEDELTA_PATTERN.match(delta)
    if match:
        parts = {k: int(v) for k, v in match.groupdict().items() if v}
        return datetime.timedelta(**parts)
    return datetime.timedelta(seconds=0)


def parse_iso(value):
    DATE_SEPARATORS = {"-", ":"}
    # date validation at speed is hard, dateutil is great but really slow, this is fast
    # but error-prone. It assumes it is a date or it really nothing like a date.
    # Making that assumption - and accepting the consequences - we can convert upto
    # three times faster than dateutil.
    #
    # valid formats (not exhaustive):
    #
    #   YYYY-MM-DD                 <- date
    #   YYYY-MM-DD HH:MM           <- date and time, no seconds
    #   YYYY-MM-DDTHH:MM           <- date and time, T separator
    #   YYYY-MM-DD HH:MM:SS        <- date and time with seconds
    #   YYYY-MM-DD HH:MM:SS.mmmm   <- date and time with milliseconds
    #
    # If the last character is a Z, we ignore it.
    # If we have +0000 notation on the end, we ignore it.
    try:
        input_type = type(value)

        if numpy:
            if input_type == numpy.datetime64:
                # this can create dates rather than datetimes, so don't return yet
                value = value.astype(datetime.datetime)
                input_type = type(value)

        if input_type == datetime.datetime:
            return value
        if input_type == datetime.date:
            return datetime.datetime.combine(value, datetime.time.min)
        if input_type in (int, float):
            return datetime.datetime.fromtimestamp(value)
        if input_type == str and 10 <= len(value) <= 33:
            if value[-1] == "Z":
                value = value[:-1]
            if "+" in value:
                value = value.split("+")[0]
                if not 10 <= len(value) <= 28:
                    return None
            val_len = len(value)
            if not value[4] in DATE_SEPARATORS or not value[7] in DATE_SEPARATORS:
                return None
            if val_len == 10:
                # YYYY-MM-DD
                return datetime.datetime(*map(int, [value[:4], value[5:7], value[8:10]]))
            if val_len >= 16:
                if not (value[10] in ("T", " ") and value[13] in DATE_SEPARATORS):
                    return False
                if val_len >= 19 and value[16] in DATE_SEPARATORS:
                    # YYYY-MM-DD HH:MM:SS
                    return datetime.datetime(
                        *map(  # type:ignore
                            int,
                            [
                                value[:4],  # YYYY
                                value[5:7],  # MM
                                value[8:10],  # DD
                                value[11:13],  # HH
                                value[14:16],  # MM
                                value[17:19],  # SS
                            ],
                        )
                    )
                if val_len == 16:
                    # YYYY-MM-DD HH:MM
                    return datetime.datetime(
                        *map(  # type:ignore
                            int,
                            [
                                value[:4],
                                value[5:7],
                                value[8:10],
                                value[11:13],
                                value[14:16],
                            ],
                        )
                    )
        return None
    except (ValueError, TypeError):
        return None


def date_range(
    start_date: Optional[Union[str, datetime.date, datetime.datetime]],
    end_date: Optional[Union[str, datetime.date, datetime.datetime]],
):
    """
    An interator over a range of dates, hour-by-hour
    """
    # if dates aren't provided, use today

    # cover the full day if we've only been given a date
    if type(end_date) == datetime.date:
        end_date = datetime.datetime.combine(end_date, datetime.time.max)
    if isinstance(end_date, str) and len(end_date) == 10:
        end_date += "T23:00:00"

    end_date = parse_iso(end_date)
    if end_date is None:
        end_date = datetime.datetime.utcnow()
    end_date = end_date.replace(minute=0, second=0, microsecond=0)  # type:ignore

    start_date = parse_iso(start_date)
    if start_date is None:
        start_date = datetime.datetime.utcnow()
    start_date = start_date.replace(minute=0, second=0, microsecond=0)  # type:ignore

    if end_date < start_date:  # type:ignore
        raise ValueError("date_range: end_date must be the same or later than the start_date ")

    for n in range(
        int((end_date - start_date).total_seconds() // SECONDS_PER_HOUR)  # type:ignore
        + 1  # type:ignore
    ):  # type:ignore
        yield start_date + datetime.timedelta(hours=n)  # type:ignore



"""
There's lots of things to time the duration of, this helps.

~~~
with Timer():
    the thing to time
~~~
"""

import time


class Timer(object):
    def __init__(self, name="test"):
        self.name = name

    def __enter__(self):
        self.start = time.time_ns()

    def __exit__(self, type, value, traceback):
        print("{} took {} seconds".format(self.name, (time.time_ns() - self.start) / 1e9))



import datetime
from typing import Any
from typing import Optional
from typing import Union

import orjson
from orso.logging import get_logger
from orso.schema import RelationSchema

from mabel.data.validator import schema_loader
from mabel.data.writers.internals.blob_writer import BlobWriter
from mabel.errors import InvalidDataSetError
from mabel.errors import MissingDependencyError
from mabel.utils import dates
from mabel.utils import paths

logger = get_logger()


class Writer:
    records = 0

    def _get_writer_date(self, date):
        # default to today if not given a date
        batch_date = datetime.datetime.utcnow()
        if isinstance(date, datetime.date):
            batch_date = date  # type:ignore
        if isinstance(date, str):
            batch_date = dates.parse_iso(date)
        return batch_date

    def __init__(
        self,
        *,
        schema: Union[RelationSchema, list, bool] = None,
        set_of_expectations: Optional[list] = None,
        format: str = "zstd",
        date: Any = None,
        partitions=["year_{yyyy}/month_{mm}/day_{dd}"],
        **kwargs,
    ):
        """
        Simple Writer provides a basic writer capability.
        """
        dataset = kwargs.get("dataset", "")

        if "BACKOUT" in dataset:
            InvalidDataSetError("BACKOUT is a reserved word and cannot be used in Dataset names")
        if dataset.endswith("/"):
            InvalidDataSetError("Dataset names cannot end with /")
        if "{" in dataset or "}" in dataset:
            InvalidDataSetError("Dataset names cannot contain { or }")
        if "%" in dataset:
            InvalidDataSetError("Dataset names cannot contain %")

        # handle transitional states - use the new features to override the legacy features
        if kwargs.get("raw_path") is not None:
            logger.warning("`raw_path` is being deprecated, set `partitions` to `None` instead.")
        if str(kwargs.get("raw_path", "")).upper() == "TRUE":
            partitions = None
        if "{date" in dataset:
            logger.warning(
                "settting the date partition in the dataset name is being deprecated, use `partitions` instead."
            )
            if "{datefolders}" in dataset:
                logger.warning("Overriding {datefolders} in dataset name")
                dataset = dataset.replace("{datefolders}", "")
                partitions = ["year_{yyyy}/month_{mm}/day_{dd}"]
            if "{datefolders_short}" in dataset:
                logger.warning("Overriding {datefolders_short} in dataset name")
                dataset = dataset.replace("{datefolders_short}", "")
                partitions = ["{yyyy}/{mm}/{dd}"]

        self.schema = schema_loader(schema)
        if not self.schema and format == "parquet":
            raise ValueError("Mabel can't write to Parquet without a schema.")

        self.expectations = None
        if set_of_expectations:
            try:
                import data_expectations as de  # type: ignore
            except:
                raise MissingDependencyError(
                    "`data_expectations` is missing, please install or include in requirements.txt"
                )
            self.expectations = de.Expectations(set_of_expectations=set_of_expectations)

        self.finalized = False
        self.batch_date = self._get_writer_date(date)

        self.dataset_template = dataset
        self.date_partitions = partitions
        if partitions:
            self.dataset_template += "/" + "/".join(partitions)
            self.date_partitions = None

        self.dataset = paths.build_path(self.dataset_template, self.batch_date)

        # add the values to kwargs
        # kwargs["raw_path"] = True  # we've just added the dates
        kwargs["format"] = format
        kwargs["dataset"] = self.dataset

        arg_dict = kwargs.copy()
        arg_dict["dataset"] = f"{self.dataset}"
        arg_dict[
            "inner_writer"
        ] = f"{arg_dict.get('inner_writer', type(None)).__name__}"  # type:ignore
        logger.debug(orjson.dumps(arg_dict))

        # add the schema to the writer - pyarrow uses this
        # add after the config has been written to the logs
        kwargs["schema"] = self.schema

        # create the writer
        self.blob_writer = BlobWriter(**kwargs)
        self.records = 0

    def append(self, record: dict):
        """
        Append a new record to the Writer

        Parameters:
            record: dictionary or pydantic.BaseModel
                The record to append to the Writer

        Returns:
            integer
                The number of records in the current blob
        """
        if "BaseModel" in str(type(record)):
            if hasattr(record, "dict"):
                record = record.dict()  # type.ignore
            if hasattr(record, "model_dump"):
                record = record.model_dump()  # type:ignore

        if self.expectations:
            import data_expectations as de  # type: ignore

            de.evaluate_record(self.expectations, record)

        if self.schema:
            self.schema.validate(record)

        self.blob_writer.append(record)
        self.records += 1

    def __del__(self):
        if hasattr(self, "finalized") and not self.finalized and self.records > 0:
            logger.error(
                f"{type(self).__name__} has not been finalized - {self.records} may have been lost, use `.finalize()` to finalize writers."
            )

    def finalize(self, **kwargs):
        self.finalized = True
        try:
            return self.blob_writer.commit()
        except Exception as e:
            logger.error(f"{type(self).__name__} failed to close pool: {type(e).__name__} - {e}")
            raise e



from mabel import DictSet

BAR_CHARS = [r" ", r"▁", r"▂", r"▃", r"▄", r"▅", r"▆", r"▇", r"█"]


class Histogram:
    def __init__(self, values, *, key: str, number_of_bins: int = 10):
        if isinstance(values, DictSet):
            values = values.collect_list(key)

        self.bins = [0] * number_of_bins
        mn = min(values)
        mx = max(values)
        interval = ((1 + mx) - mn) / (number_of_bins - 1)

        for v in values:
            self.bins[int((v - mn) / interval)] += 1

    def __repr__(self):
        """
        Draws a pre-binned set off histogram data
        """
        mx = max(self.bins)
        bar_height = mx / 8
        if bar_height == 0:
            return ">" + " " * len(self.bins) + "<"

        histogram = ""
        for value in self.bins:
            if value == 0:
                histogram += BAR_CHARS[0]
            else:
                height = int(value / bar_height)
                histogram += BAR_CHARS[height]

        return f">{histogram}<"



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data.writers.internals.writer_pool import WriterPool
from mabel.adapters.null import NullWriter
from rich import traceback

traceback.install()


def test_writer_pool_lru():
    pool = WriterPool(3, inner_writer=NullWriter, schema=["record"])
    pool.get_writer("beatle/john.txt")
    pool.get_writer("beatle/paul.txt")
    pool.get_writer("beatle/george.txt")

    # with three items, nothing should be on the eviction list
    assert len(pool.nominate_writers_to_evict()) == 0
    pool.get_writer("beatle/ringo.txt")

    # we've added four items but not evicted any, should have four items
    assert len(pool.writers) == 4
    # four is one more than three
    to_evict = pool.nominate_writers_to_evict()
    assert len(to_evict) == 1
    # john is the least recently used
    assert to_evict[0] == "beatle/john.txt"

    pool.get_writer("beatle/john.txt")

    # we accessed john, paul us now the LRU
    to_evict = pool.nominate_writers_to_evict()
    assert len(to_evict) == 1
    assert to_evict[0] == "beatle/paul.txt"


def test_writer_add_and_remove():
    pool = WriterPool(3, inner_writer=NullWriter, schema=["record"])
    pool.get_writer("beatle/john.txt")
    pool.get_writer("beatle/paul.txt")

    # we should have two writers in the pool
    assert len(pool.writers) == 2

    pool.remove_writer("beatle/john.txt")

    # we removed one so we should have only paul in the pool
    assert len(pool.writers) == 1
    assert pool.writers[0].get("identity") == "beatle/paul.txt"

    pool.get_writer("beatle/george.txt")

    # george should be in the pool with paul
    assert len(pool.writers) == 2


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""
Google Cloud Storage Reader
"""
import os

from mabel.data.readers.internals.base_inner_reader import BaseInnerReader
from mabel.errors import MissingDependencyError
from mabel.utils import paths

try:
    from google.auth.credentials import AnonymousCredentials  # type:ignore
    from google.cloud import storage  # type:ignore

    google_cloud_storage_installed = True
except ImportError:  # pragma: no cover
    google_cloud_storage_installed = False


class GoogleCloudStorageReader(BaseInnerReader):
    def __init__(self, credentials=None, **kwargs):
        if not google_cloud_storage_installed:  # pragma: no cover
            raise MissingDependencyError(
                "`google-cloud-storage` is missing, please install or include in requirements.txt"
            )

        super().__init__(**kwargs)
        self.credentials = credentials

    def get_blob_bytes(self, blob_name):
        bucket, object_path, name, extension = paths.get_parts(blob_name)
        blob = get_blob(
            bucket=bucket,
            blob_name=object_path + name + extension,
        )
        stream = blob.download_as_bytes()
        return stream

    def get_blobs_at_path(self, path):
        bucket, object_path, name, extension = paths.get_parts(path)

        # this means we're testing
        if os.environ.get("STORAGE_EMULATOR_HOST") is not None:
            client = storage.Client(credentials=AnonymousCredentials())
        else:  # pragma: no cover
            client = storage.Client()

        gcs_bucket = client.get_bucket(bucket)
        blobs = list(client.list_blobs(bucket_or_name=gcs_bucket, prefix=object_path))

        yield from [bucket + "/" + blob.name for blob in blobs if not blob.name.endswith("/")]


def get_blob(bucket: str = None, blob_name: str = None):
    # this means we're testing
    if os.environ.get("STORAGE_EMULATOR_HOST") is not None:
        client = storage.Client(credentials=AnonymousCredentials())
    else:  # pragma: no cover
        client = storage.Client()

    gcs_bucket = client.get_bucket(bucket)
    blob = gcs_bucket.get_blob(blob_name)
    return blob



import sys
import os
import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.utils import common
from rich import traceback

traceback.install()


def test_utils_config():
    config = {}
    with pytest.raises((IndexError)):
        common.build_context(config_file="not existent config file")

    with pytest.raises((ValueError)):
        common.build_context(config_file="tests/data/invalid.config")


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""
StorageClassDisk is a helper class for persisting DictSets locally, it is the backend
for the DISK variation of the STORAGE CLASSES.

The Reader and Writer are pretty fast, the bottleneck is the parsing and serialization
of JSON data - this accounts for over 50% of the read/write times.
"""
import atexit
import mmap
import os
from tempfile import NamedTemporaryFile

from mabel.utils.paths import silent_remove

from . import BaseStorageClass

BUFFER_SIZE = 16 * 1024 * 1024  # 16Mb


class StorageClassDisk(BaseStorageClass):
    """
    This provides the reader for the DISK variation of STORAGE.
    """

    def __init__(self, iterator):
        self.inner_reader = None
        self.length = -1

        self.file = NamedTemporaryFile(prefix="mabel-dictset").name
        atexit.register(silent_remove, filename=self.file)

        buffer = bytearray()
        with open(self.file, "wb") as f:
            for self.length, row in enumerate(iterator):
                # there is a penalty for using this object
                if hasattr(row, "mini"):
                    buffer.extend(row.mini + b"\n")
                else:
                    buffer.extend(self.dump_json(row) + b"\n")
                if len(buffer) > (BUFFER_SIZE):
                    f.write(buffer)
                    buffer = bytearray()
            if len(buffer) > 0:
                f.write(buffer)
            f.flush()

        self.length += 1
        self.iterator = None

    def _read_file(self):
        """
        MMAP is by far the fastest way to read files in Python.
        """
        with open(self.file, mode="rb") as file_obj:
            with mmap.mmap(file_obj.fileno(), length=0, access=mmap.ACCESS_READ) as mmap_obj:
                line = mmap_obj.readline()
                while line:
                    yield line
                    line = mmap_obj.readline()

    def _inner_reader(self, *locations):
        if locations:
            max_location = max(locations)
            min_location = min(locations)

            reader = self._read_file()

            for i in range(min_location):
                next(reader)

            for i, line in enumerate(reader, min_location):
                if i in locations:
                    yield self.parse_json(line)
                    if i == max_location:
                        return
        else:
            for line in self._read_file():
                yield self.parse_json(line)

    def __iter__(self):
        self.iterator = iter(self._inner_reader())
        return self.iterator

    def __next__(self):
        if not self.iterator:
            self.iterator = iter(self._inner_reader())
        return next(self.iterator)

    def __len__(self):
        return self.length

    def __del__(self):
        try:
            os.remove(self.file)
        except:  # nosec
            pass



import orjson

pass_thru = lambda x: x

json = orjson.loads


def pass_thru_block(ds):
    """each blob is read as a block"""
    if isinstance(ds, str):
        return ds
    return "\n".join([r for r in ds])  # pragma: no cover


def xml(ds):
    from ...internals import xmler

    return xmler.parse(ds)



import shutil
import datetime
import os
import sys
import glob

from pyarrow import parquet

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.disk import DiskReader, DiskWriter
from mabel.data import BatchWriter
from mabel.data import Reader
from mabel.data.internals.dictset import STORAGE_CLASS
from mabel.utils import entropy
from rich import traceback

traceback.install()


def do_writer():
    w = BatchWriter(
        inner_writer=DiskWriter,
        dataset="_temp",
        date=datetime.datetime.utcnow().date(),
        format="parquet",
        schema={
            "fields": [
                {"name": "name", "type": "VARCHAR"},
                {"name": "alias", "type": "VARCHAR"},
                {"name": "season", "type": "INTEGER"},
                {"name": "recurring", "type": "BOOLEAN"},
                {"name": "list_test", "type": "ARRAY"},
                {"name": "dict_test", "type": "STRUCT"},
            ]
        },
    )
    for i in range(int(5e5)):
        w.append(
            {
                "name": None,
                "alias": entropy.random_string(256),
                "season": None,
                "recurring": None,
                "list_test": [],
                "dict_test": {"one": "1"},
            }
        )
    for i in range(int(1e5)):
        w.append(
            {
                "name": "Barney Stinson",
                "alias": "Lorenzo Von Matterhorn",
                "season": 4,
                "recurring": False,
                "list_test": ["a"],
                "dict_test": {"a": "b"},
            }
        )
        w.append(
            {
                "name": "Laszlo Cravensworth",
                "alias": "Jackie Daytona",
                "season": 3,
                "recurring": True,
                "list_test": ["ai", "bee", "see", "dee", "eff", "ghee"],
                "dict_test": {"alpha": "bet", "a": "gamma"},
            }
        )
    w.finalize()


def test_reader_writer_parquet_normalization():
    do_writer()

    parquet_files = glob.glob("_temp/**/*.parquet", recursive=True)

    for parquet_file in parquet_files:
        table = parquet.read_table(parquet_file)
        print(table.schema.names, table.schema.types)

    assert len(parquet_files) > 0, parquet_files

    c = glob.glob("_temp/**/*.complete", recursive=True)
    len(c) == 0, c

    print("written")
    parq = Reader(inner_reader=DiskReader, dataset="_temp", persistence=STORAGE_CLASS.MEMORY)
    records = parq.count()
    recurring = parq.collect_list("recurring")
    aliases = [
        isinstance(a, datetime.datetime) for a in parq.collect_list("alias") if a is not None
    ]
    shutil.rmtree("_temp", ignore_errors=True)
    assert records == 700000, records
    assert recurring.count(True) == 100000, recurring.count(True)
    assert recurring.count(None) == 500000, recurring.count(None)
    assert all(a is not None for a in aliases), all(a is not None for a in aliases)


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""
Written following a bug discovered in how the Reader determines the data range.
"""
import datetime
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data.readers.internals.base_inner_reader import BaseInnerReader
from rich import traceback
from mabel import Reader
from mabel.adapters.disk import DiskReader

traceback.install()


class EmptyReader(BaseInnerReader):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_blobs_at_path(self):
        pass

    def get_blob_chunk(self):
        pass

    def get_blob_bytes(self):
        pass


START_DATE = datetime.datetime(2000, 1, 1)
END_DATE = datetime.datetime(2010, 6, 15)
TODAY = datetime.datetime.utcnow().replace(minute=0, second=0, microsecond=0)


def test_start_date_only():
    subject = EmptyReader(dataset="", start_date=START_DATE)
    assert subject.start_date == START_DATE, subject.start_date
    assert subject.end_date == TODAY, subject.end_date


def test_end_date_only():
    subject = EmptyReader(dataset="", end_date=END_DATE)
    # the values are reversed to make a valid range of start < end
    assert subject.start_date == END_DATE, subject.start_date
    assert subject.end_date == TODAY, subject.end_date


def test_start_and_end_dates():
    subject = EmptyReader(dataset="", start_date=START_DATE, end_date=END_DATE)
    assert subject.start_date == START_DATE
    assert subject.end_date == END_DATE


def test_dates_as_string():
    subject = EmptyReader(dataset="", start_date="2000-01-01", end_date="2010-06-15")
    assert subject.start_date == START_DATE, subject.start_date
    assert subject.end_date == END_DATE


def test_reading_across_dates():
    subject = Reader(
        inner_reader=DiskReader,
        dataset="tests/data/dated",
        start_date="2020-02-03",
        end_date="2020-02-04",
    )
    record_count = len(list(subject))
    assert record_count == 50, record_count


def test_short_dates():
    subject = Reader(
        inner_reader=DiskReader,
        dataset="tests/data/dated",
        partitions=["{yyyy}/{mm}/{dd}"],
        start_date="1979-08-23",
        end_date="1979-08-23",
    )
    assert len(list(subject)) == 25


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import os
import sys

# assert False, "test records over the partition size are saved"
# assert False, "test the partitions are respected"

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.null import NullWriter
from mabel.data import Writer, BatchWriter
from rich import traceback

traceback.install()

from mabel.data.writers.internals.blob_writer import BLOB_SIZE


def test_null_writer():
    # none of these should do anything
    nw = Writer(inner_writer=NullWriter, schema=["blob"], dataset="bucket/path")
    for i in range(1):
        nw.append({"blob": ["2"] * BLOB_SIZE})
    res = nw.finalize()
    assert res.startswith("NullWriter"), res


def test_hourly_partitions():
    nw = BatchWriter(
        inner_writer=NullWriter,
        dataset="bucket/path",
        schema=["@"],
        partitions=["year_{yyyy}/month_{mm}/day_{dd}/by_hour/hour={HH}"],
    )
    for i in range(1):
        nw.append({"@": [" "] * BLOB_SIZE})
    res = nw.finalize()
    assert "by_hour/hour=" in res


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



from ...data.readers.internals.base_inner_reader import BaseInnerReader


class NullReader(BaseInnerReader):
    RULES = [{"name": "data", "required": True}]

    def __init__(self, **kwargs):
        """
        Null reader
        """
        super().__init__(**kwargs)
        self._data = kwargs.get("data", [])

    def get_blobs_at_path(self, path):
        return [f"{path}/null.jsonl"]

    def get_blob_bytes(self, blob_name: str) -> bytes:
        import orjson

        return b"\n".join(map(orjson.dumps, self._data))



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
import pytest
from mabel.data.readers.internals.sql_reader import InvalidSqlError, SqlParser
from rich import traceback

traceback.install()


def test_parser():
    # fmt:off
    STATEMENTS = [
        {"SQL": "SELECT * FROM TABLE", "select": "*", "from": "TABLE"},
        {"SQL": "SELECT DATE(dob) FROM TABLE", "select": "DATE(dob)", "from": "TABLE"},
        {"SQL": "SELECT * --everything \n FROM TABLE", "select": "*", "from": "TABLE"},
        {"SQL": "/*this\nis\na\ncomment*/SELECT * --everything \n FROM TABLE", "select": "*", "from": "TABLE"},
        {"SQL": "SELECT * \n FROM TABLE \n/* 2 this\nis\na\ncomment */", "select": "*", "from": "TABLE"},
        {"SQL": "/*this\nis\na\ncomment*/SELECT * --everything \n FROM TABLE/*this\nis\na second\ncomment*/", "select": "*", "from": "TABLE"},
        {"SQL": "SELECT * --everything \nFROM TABLE --this table", "select": "*", "from": "TABLE"},
        {"SQL": "SELECT value FROM TABLE", "select": "value", "from": "TABLE"},
        {"SQL": "SELECT value1, value2 FROM TABLE","select": "value1, value2","from": "TABLE",},
        {"SQL": "SELECT\n\tvalue\nFROM\nTABLE\nWHERE\n\tvalue == 1","select": "value","from": "TABLE","where": "value == 1",},
        {"SQL": "SELECT COUNT(*) FROM TABLE WHERE value == 1 GROUP BY value LIMIT 3","select": "COUNT(*)", "from": "TABLE","where": "value == 1","group_by": "value","limit": 3,},
        {"SQL": "SELECT \n    MAX(cve.CVE_data_meta.ID),\n    MIN(cve.CVE_data_meta.ID),\n    COUNT(cve.CVE_data_meta.ID) \nFROM mabel_data.RAW.NVD.CVE_LIST GROUP BY cve.CVE_data_meta.ASSIGNER","select": "MAX(cve.CVE_data_meta.ID), MIN(cve.CVE_data_meta.ID), COUNT(cve.CVE_data_meta.ID)","from": "mabel_data/RAW/NVD/CVE_LIST","group_by": "cve.CVE_data_meta.ASSIGNER"},
        {"SQL": "SELECT MAX(AGE), MIN(AGE), HASH(AGE) FROM TABLE", "select": "MAX(AGE), MIN(AGE), HASH(AGE)", "from": "TABLE"},
        {"SQL": "SELECT DISTINCT * FROM TABLE", "select": "*", "from": "TABLE"},
        {"SQL": "SELECT * FROM TABLE HAVING HASH(AGE) > 100", "select": "*", "from": "TABLE"},
        {"SQL": "SELECT * FROM TABLE ORDER BY apples", "select": "*", "from": "TABLE"},
        {"SQL": "SELECT * FROM TABLE ORDER BY apples DESC", "select": "*", "from": "TABLE"},
    ]
    # fmt:on

    for statement in STATEMENTS:
        print(statement["SQL"])
        parsed = SqlParser(statement["SQL"])
        assert parsed.select_expression == statement.get(
            "select"
        ), f"SELECT {statement.get('select')}: {parsed}"
        assert parsed.dataset == statement.get("from"), f"FROM: {parsed}"
        assert parsed.where_expression == statement.get("where"), f"WHERE: {parsed}"
        assert parsed.group_by == statement.get("group_by"), f"GROUP BY: {parsed}"
        assert parsed.limit == statement.get("limit"), f"LIMIT: {parsed}"


def test_invalid_sql():
    STATEMENTS = [
        # "SELECT * FROM TABLE WHERE value == 1 GROUP BY value LIMIT 3",   <-- currently parses but raises error when run
        "SELECT *",
    ]

    for statement in STATEMENTS:
        # print(statement)
        with pytest.raises(Exception):
            parsed = SqlParser(statement)


def test_from_validator():
    assert SqlParser.validate_dataset(None, "table")
    assert SqlParser.validate_dataset(None, "data.set/is_valid")

    with pytest.raises(InvalidSqlError):
        SqlParser.validate_dataset(None, "1table")
    with pytest.raises(InvalidSqlError):
        SqlParser.validate_dataset(None, ".table")
    with pytest.raises(InvalidSqlError):
        SqlParser.validate_dataset(None, "this//isnotvalid")
    with pytest.raises(InvalidSqlError):
        SqlParser.validate_dataset(None, "this*is*not*okay")
    with pytest.raises(InvalidSqlError):
        SqlParser.validate_dataset(None, "this--is--not--okay")


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.disk import DiskReader
from mabel.data.internals.dnf_filters import DnfFilters
from mabel.data.internals.dictset import STORAGE_CLASS
from mabel.data import Reader
from rich import traceback

traceback.install()

# fmt: off
TEST_DATA = [
    { "name": "Sirius Black", "age": 40, "dob": "1970-01-02", "gender": "male", "affiliations": ['OotP'] },
    { "name": "Harry Potter", "age": 11, "dob": "1999-07-30", "gender": "male", "affiliations": ['Dumbledores Army', 'Griffindor'] },
    { "name": "Hermione Grainger", "age": 10, "dob": "1999-12-14", "gender": "female", "affiliations": ['Dumbledores Army', 'Griffindor', 'MoM'] },
    { "name": "Fleur Isabelle Delacour", "age": 11, "dob": "1999-02-08", "gender": "female" },
    { "name": "James Potter", "age": 40, "dob": "1971-12-30", "gender": "male" },
    { "name": "James Potter", "age": 0, "dob": "2010-12-30", "gender": "male" }
]
# fmt: on


def test_reader_filters_no_filter():
    """ensure the reader filter is working as expected"""
    r = Reader(inner_reader=DiskReader, dataset="tests/data/tweets/", raw_path=True)
    for index, item in enumerate(r):
        pass
    assert index == 49, index


def test_reader_filters_single_filter():
    """ensure the reader filter is working as expected"""
    r = Reader(
        inner_reader=DiskReader,
        dataset="tests/data/tweets/",
        raw_path=True,
        filters="username == 'NBCNews'",
        persistence=STORAGE_CLASS.MEMORY,
    )
    assert r.count() == 44, r.count()


def test_reader_filters_multiple_filter():
    """ensure the reader filter is working as expected"""
    r = Reader(
        inner_reader=DiskReader,
        dataset="tests/data/tweets/",
        raw_path=True,
        filters="username = 'NBCNews' and timestamp >= '2020-01-12T07:11:04'",
        persistence=STORAGE_CLASS.MEMORY,
    )
    assert r.count() == 34, r


def test_filters():
    d = DnfFilters(filters=[("age", "==", 11), ("gender", "in", ("a", "b", "male"))])
    assert len(list(d.filter_dictset(TEST_DATA))) == 1


def test_empty_filters():
    d = DnfFilters(filters=None)
    assert len(list(d.filter_dictset(TEST_DATA))) == 6


def test_like_filters():
    d = DnfFilters(filters=[("dob", "like", "%-12-%")])
    assert len(list(d.filter_dictset(TEST_DATA))) == 3


def test_combined_filters():
    # just a tuple
    filter01 = DnfFilters(("name", "==", "Harry Potter"))
    # a tuple in a list
    filter02 = DnfFilters([("name", "==", "Harry Potter")])
    # ANDed conditions
    filter03 = DnfFilters(
        [
            ("name", "like", "%otter%"),
            ("gender", "==", "male"),
            ("age", "<=", 15),
            ("age", ">=", 5),
        ]
    )
    # ANDed conditions - case insensitive LIKE
    filter03a = DnfFilters(
        [
            ("name", "like", "%pOTTER%"),
            ("gender", "==", "male"),
            ("age", "<=", 15),
            ("age", ">=", 5),
        ]
    )
    # ORed conditions
    filter04 = DnfFilters([[("name", "==", "Harry Potter")], [("name", "==", "Hermione Grainger")]])
    # no conditions
    filter05 = DnfFilters()
    # IN conditions
    filter06 = DnfFilters(("name", "in", ["Fleur Isabelle Delacour", "Hermione Grainger"]))
    # contains conditions - on lists
    filter07 = DnfFilters(("affiliations", "contains", "Griffindor"))
    # contains conditions
    filter08 = DnfFilters(("name", "contains", "Isabelle"))

    assert len([a for a in filter01.filter_dictset(TEST_DATA)]) == 1
    assert len([a for a in filter02.filter_dictset(TEST_DATA)]) == 1
    assert len([a for a in filter03.filter_dictset(TEST_DATA)]) == 1
    assert len([a for a in filter03a.filter_dictset(TEST_DATA)]) == 1
    assert len([a for a in filter04.filter_dictset(TEST_DATA)]) == 2
    assert len([a for a in filter05.filter_dictset(TEST_DATA)]) == 6
    assert len([a for a in filter06.filter_dictset(TEST_DATA)]) == 2
    assert len([a for a in filter07.filter_dictset(TEST_DATA)]) == 2
    assert len([a for a in filter08.filter_dictset(TEST_DATA)]) == 1


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import datetime
import threading
from typing import Optional
from typing import Union

import orso
from orso.logging import get_logger
from orso.schema import RelationSchema

from mabel.data.validator import schema_loader
from mabel.errors import MissingDependencyError
from mabel.utils import dates

logger = get_logger()

BATCH_SIZE: int = 100


class DatabaseWriter:
    records = 0

    def _get_writer_date(self, date):
        # default to today if not given a date
        batch_date = datetime.datetime.utcnow()
        if isinstance(date, datetime.date):
            batch_date = date  # type:ignore
        if isinstance(date, str):
            batch_date = dates.parse_iso(date)
        return batch_date

    def __init__(
        self,
        *,
        schema: Union[RelationSchema, list] = None,
        set_of_expectations: Optional[list] = None,
        inner_writer=None,
        **kwargs,
    ):
        """
        Simple Writer provides a basic writer capability.
        """
        self.dataset = kwargs.get("dataset", "")
        self.schema = schema_loader(schema)
        if not self.schema:
            raise ValueError("Cannot use Database Writer without a schema")

        self.expectations = None
        if set_of_expectations:
            try:
                import data_expectations as de  # type: ignore
            except:
                raise MissingDependencyError(
                    "`data_expectations` is missing, please install or include in requirements.txt"
                )
            self.expectations = de.Expectations(set_of_expectations=set_of_expectations)

        self.finalized = False

        kwargs["dataset"] = self.dataset

        arg_dict = kwargs.copy()
        arg_dict["dataset"] = f"{self.dataset}"
        arg_dict["inner_writer"] = f"{inner_writer.__name__}"  # type:ignore
        logger.debug([f"{k}={v!r}" for k, v in arg_dict.items()])

        # add the schema to the writer - pyarrow uses this
        # add after the config has been written to the logs
        kwargs["schema"] = self.schema

        # create the writer
        self.inner_writer = inner_writer(**kwargs)
        self.records = 0

        self.wal = orso.DataFrame(rows=[], schema=self.schema)
        self.records_in_buffer = 0

    def append(self, record: dict):
        """
        Append a new record to the Writer

        Parameters:
            record: dictionary or pydantic.BaseModel
                The record to append to the Writer

        Returns:
            integer
                The number of records in the current blob
        """
        if "BaseModel" in str(type(record)):
            if hasattr(record, "dict"):
                record = record.dict()  # type.ignore
            if hasattr(record, "model_dump"):
                record = record.model_dump()  # type:ignore

        if self.expectations:
            import data_expectations as de  # type: ignore

            de.evaluate_record(self.expectations, record)

        self.wal.append(record)
        self.records_in_buffer += 1
        self.records += 1
        # if this write would exceed the blob size, close it
        if self.wal.rowcount >= BATCH_SIZE:
            self.commit()
            self.wal = orso.DataFrame(rows=[], schema=self.schema)
            self.records_in_buffer = 0

        return self.records_in_buffer

    def __del__(self):
        if hasattr(self, "finalized") and not self.finalized and self.records > 0:
            logger.error(
                f"{type(self).__name__} has not been finalized - {self.records} may have been lost, use `.finalize()` to finalize writers."
            )

    def commit(self):
        if len(self.wal) > 0:
            lock = threading.Lock()

            try:
                lock.acquire(blocking=True, timeout=10)

                self.inner_writer.commit(self.wal)

                get_logger().debug(
                    {
                        "action": "commit",
                        "format": self.inner_writer.__class__.__name__,
                        "records": len(self.wal),
                    }
                )
            finally:
                lock.release()

            self.wal = orso.DataFrame(rows=[], schema=self.schema)
            self.records_in_buffer = 0

    def finalize(self):
        # flush the WAL into the database
        self.commit()

        lock = threading.Lock()
        try:
            lock.acquire(blocking=True, timeout=10)

            self.inner_writer.finalize()

            get_logger().debug(
                {"action": "finalize", "format": self.inner_writer.__class__.__name__}
            )
        except Exception as e:
            get_logger().error(
                f"Failed to finalize database write with {self.records} to {self.dataset}"
            )
            raise Exception from e
        finally:
            lock.release()

        return self.records

    def __del__(self):
        # this should never be relied on to save data
        self.commit()
        self.finalize()



import datetime
import io
import sys
import threading

import orso
import zstandard
from orso.logging import get_logger

from mabel.data.validator import Schema
from mabel.errors import MissingDependencyError

BLOB_SIZE = 64 * 1024 * 1024  # 64Mb, 16 files per gigabyte
SUPPORTED_FORMATS_ALGORITHMS = ("jsonl", "zstd", "parquet", "text", "flat")


def get_size(obj, seen=None):
    """
    Recursively approximate the size of objects.
    We don't know the actual size until we save, so we approximate the size based
    on some rules - this will be wrong due to RLE, headers, precision and other
    factors.
    """
    obj_type = type(obj)
    obj_id = id(obj)

    if seen is None:
        seen = set()

    if obj_id in seen:
        return 0

    if obj_type == int or obj_type == float:
        size = 6
    elif obj_type == bool:
        size = 1
    elif obj_type in {str, bytes, bytearray}:
        size = len(obj) + 4
    elif obj is None:  # NoneType is not directly accessible
        size = 1
    elif obj_type == datetime.datetime:
        size = 8
    else:
        size = sys.getsizeof(obj)

    # Mark as seen before entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)

    if obj_type == dict:
        size += sum(get_size(v, seen) for v in obj.values()) + 8
    elif hasattr(obj, "__dict__"):
        size += get_size(obj.__dict__, seen) + 8
    elif hasattr(obj, "__iter__") and obj_type not in {str, bytes, bytearray}:
        size += sum(get_size(i, seen) for i in obj) + 8

    return size


class BlobWriter(object):
    wal = []

    def __init__(
        self,
        *,  # force params to be named
        inner_writer=None,  # type:ignore
        blob_size: int = BLOB_SIZE,
        format: str = "zstd",
        schema: Schema = None,
        **kwargs,
    ):
        self.format = format
        self.maximum_blob_size = blob_size

        if format not in SUPPORTED_FORMATS_ALGORITHMS:
            raise ValueError(
                f"Invalid format `{format}`, valid options are {SUPPORTED_FORMATS_ALGORITHMS}"
            )

        kwargs["format"] = format
        self.inner_writer = inner_writer(**kwargs)  # type:ignore

        self.schema = None
        if isinstance(schema, (list, dict)):
            schema = Schema(schema)
        if isinstance(schema, Schema):
            self.schema = schema
        if self.schema is None:
            self.schema = []
            raise MissingDependencyError("Writers require a Schema")
        self.schema = self.schema.schema

        self.wal = orso.DataFrame(rows=[], schema=self.schema)
        self.records_in_buffer = 0
        self.byte_count = 5120

    def append(self, record: dict):
        self.wal.append(record)
        self.records_in_buffer += 1
        self.byte_count += get_size(record)
        # if this write would exceed the blob size, close it
        if self.byte_count >= self.maximum_blob_size:
            self.commit()
            self.wal = orso.DataFrame(rows=[], schema=self.schema)
            self.records_in_buffer = 0
            self.byte_count = 5120

        return self.records_in_buffer

    def commit(self):
        committed_blob_name = ""

        if len(self.wal) > 0:
            lock = threading.Lock()

            try:
                lock.acquire(blocking=True, timeout=10)

                if self.format == "parquet":
                    try:
                        import pyarrow.parquet
                    except ImportError:
                        raise MissingDependencyError(
                            "`pyarrow` missing, please install or include in `requirements.txt`."
                        )
                    pytable = self.wal.arrow()

                    tempfile = io.BytesIO()
                    pyarrow.parquet.write_table(pytable, where=tempfile, compression="zstd")

                    tempfile.seek(0)
                    buffer = tempfile.read()

                else:
                    buffer = b"\n".join(r.as_json for r in self.wal) + b"\n"
                    if self.format == "zstd":
                        # zstandard is an non-optional installed dependency
                        buffer = zstandard.compress(buffer)

                committed_blob_name = self.inner_writer.commit(
                    byte_data=bytes(buffer), override_blob_name=None
                )

                if "BACKOUT" in committed_blob_name:
                    get_logger().warning(
                        f"{self.records_in_buffer:n} failed records written to BACKOUT partition `{committed_blob_name}`"
                    )
                get_logger().debug(
                    {
                        "format": self.format,
                        "committed_blob": committed_blob_name,
                        "records": len(buffer)
                        if self.format == "parquet"
                        else self.records_in_buffer,
                    }
                )
            finally:
                lock.release()

        self.wal = orso.DataFrame(rows=[], schema=self.schema)
        self.records_in_buffer = 0
        self.byte_count = 5120
        return committed_blob_name

    def __del__(self):
        # this should never be relied on to save data
        self.commit()



from .base_storage_class import BaseStorageClass
from .storage_class_compressed_memory import StorageClassCompressedMemory
from .storage_class_disk import StorageClassDisk
from .storage_class_memory import StorageClassMemory



from helpers import Mock
from rich import traceback

traceback.install()


def test_mock():
    mock = Mock({"name": "I'm a Mock", "child": Mock({"name": "My parent is a Mock"})}, ["name"])

    assert mock.name == "I'm a Mock"
    assert mock.rank() == "rank"
    assert isinstance(mock.child(), Mock)
    assert mock.child().name() == "My parent is a Mock"


if __name__ == "__main__":  # pragma: no cover
    from helpers.runner import run_tests

    run_tests()



import datetime
import os
from typing import Any

import orjson
from orso.logging import get_logger

from mabel.data.writers.internals.blob_writer import BlobWriter
from mabel.data.writers.writer import Writer
from mabel.utils import paths

logger = get_logger()


class BatchWriter(Writer):
    """
    Extend the functionality of the Writer to better support batch data
    """

    def __init__(
        self,
        *,
        dataset: str,
        format: str = "zstd",
        date: Any = None,
        partitions=["year_{yyyy}/month_{mm}/day_{dd}"],
        frame_id: str = None,
        metadata: dict = None,
        always_complete: bool = False,
        **kwargs,
    ):
        """
        The batch data writer to writes data records into blobs. Batches
        are written into timestamped folders called Partitions.

        Parameters:
            dataset: string (optional)
                The name of the dataset - this is used to map to a path
            schema: mabel.validator.Schema (optional)
                Schema used to test records for conformity, default is no
                schema and therefore no validation
            format: string (optional)
                - text: raw text lines
                - jsonl: raw json lines
                - flat: flattened json records in json lines
                - lzma: lzma compressed json lines
                - zstd: zstandard compressed json lines (default)
                - parquet: Apache Parquet
            date: date or string (optional)
                A date, a string representation of a date to use for
                creating the dataset. The default is today's date
            blob_size: integer (optional)
                The maximum size of blobs, the default is 64Mb
            inner_writer: BaseWriter (optional)
                The component used to commit data, the default writer is the
                NullWriter
            frame_id: string (optional)

            raw_path: boolean (optional)
                Don't automatically add any date parts to dataset names
            index_on: collection (optional)
                Index on these columns, the default is to not index
            metadata: dict (optional)
                data to write into the frame.complete file
            always_complete: bool (optional)
                Write a frame.complete even if zero records are written.

        Note:
            Different inner_writers may take or require additional parameters.
        """
        # because jobs can be split, allow a frame_id to be passed in
        if not frame_id:
            # as_ats are in UTC
            as_at = datetime.datetime.utcnow().strftime("as_at_%Y%m%d-%H%M%S")
            self.batch_date = self._get_writer_date(date)

            # we now use partitions
            # if "{date" not in dataset and not kwargs.get("raw_path", False):
            #    dataset += "/{datefolders}"
            if partitions:
                dataset += "/" + "/".join(partitions)
            frame_id = paths.build_path(
                dataset + "/" + as_at,  # type:ignore
                self.batch_date,
            )

        if metadata:
            self.metadata = metadata
        else:
            self.metadata = {}

        # kwargs["raw_path"] = True  # we've just added the dates
        kwargs["format"] = format
        kwargs["dataset"] = frame_id
        kwargs["partitions"] = partitions

        super().__init__(**kwargs)

        self.dataset = frame_id
        kwargs["schema"] = self.schema

        # create the writer
        self.blob_writer = BlobWriter(**kwargs)

        # in finalize is called more than once - it can be called with the
        # error cleared from the stack.
        self.seen_failures = False

        self.always_complete = always_complete

    def finalize(self, **kwargs):
        final = super().finalize()

        has_failure = bool(kwargs.get("has_failure", False))

        if has_failure:
            self.seen_failures = True
            get_logger().debug(f"Error found in the stack, not marking frame as complete.")
            return -1

        if self.seen_failures:
            get_logger().debug(
                f"Error previously seen in the stack, not marking frame as complete."
            )
            return -1

        if self.records == 0 and not self.always_complete:
            get_logger().warning(
                f"No records written, and 'always_complete' not set, so not marking frame as complete."
            )
            return -1

        completion_path = self.blob_writer.inner_writer.filename
        completion_path = os.path.split(completion_path)[0] + "/frame.complete"
        self.metadata["records"] = self.records
        if self.schema:
            if isinstance(self.schema, dict):
                self.metadata["schema"] = self.schema
            elif hasattr(self.schema, "definition"):
                self.metadata["schema"] = self.schema.definition
        flag = self.blob_writer.inner_writer.commit(
            byte_data=orjson.dumps(self.metadata, option=orjson.OPT_INDENT_2),
            override_blob_name=completion_path,
        )
        get_logger().debug(f"Frame completion file `{flag}` written")
        return final



import os
import sys
import time

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.adapters.null import NullWriter
from mabel.data import StreamWriter

from mabel.logging import get_logger

logger = get_logger()


def test_writer_flood():
    logger.setLevel(5)
    w = StreamWriter(
        dataset="path/{identity}/path",
        inner_writer=NullWriter,
        idle_timeout_seconds=1,
        schema=["identity", "cycle"],
    )

    for i in range(1000):
        record = {"identity": str(os.urandom(1)), "cycle": i}
        time.sleep(0.001)
        w.append(record)

    logger.info("FINISHED WRITING")

    print(len(w.writer_pool.writers))

    w.finalize()


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import os
import sys
from datetime import timedelta

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.utils.dates import parse_delta


def test_day_deltas():
    assert parse_delta("3d") == timedelta(days=3)
    assert parse_delta("-3d") == timedelta(days=-3)
    assert parse_delta("-37D") == timedelta(days=-37)


def test_hours_deltas():
    assert parse_delta("18h") == timedelta(hours=18)
    assert parse_delta("-5h") == timedelta(hours=-5)
    assert parse_delta("11H") == timedelta(hours=11)


def test_minute_deltas():
    assert parse_delta("129m") == timedelta(minutes=129)
    assert parse_delta("-68m") == timedelta(minutes=-68)
    assert parse_delta("12M") == timedelta(minutes=12)


def test_second_deltas():
    assert parse_delta("300s") == timedelta(seconds=300)
    assert parse_delta("-5s") == timedelta(seconds=-5)
    assert parse_delta("23S") == timedelta(seconds=23)


def test_combined_deltas():
    assert parse_delta("3d5h") == timedelta(days=3, hours=5)
    assert parse_delta("-3d-5h") == timedelta(days=-3, hours=-5)
    assert parse_delta("13d4h19m") == timedelta(days=13, hours=4, minutes=19)

    assert parse_delta("13d19m") == timedelta(days=13, minutes=19)
    assert parse_delta("-13d-19m") == timedelta(days=-13, minutes=-19)
    assert parse_delta("-13d19m") == timedelta(days=-13, minutes=19)
    assert parse_delta("13d-19m") == timedelta(days=13, minutes=-19)

    assert parse_delta("4h19m") == timedelta(hours=4, minutes=19)
    assert parse_delta("-4h-19m") == timedelta(hours=-4, minutes=-19)
    assert parse_delta("-4h19m") == timedelta(hours=-4, minutes=19)
    assert parse_delta("4h-19m") == timedelta(hours=4, minutes=-19)


def test_invalid_scenarios():
    assert parse_delta("") == timedelta(seconds=0), parse_delta("")
    assert parse_delta("five seconds") == timedelta(seconds=-0), parse_delta("five seconds")
    assert parse_delta("tjyhrsdy") == timedelta(seconds=0), parse_delta("tjyhrsdy")
    assert parse_delta("1m") == timedelta(seconds=60), parse_delta("1m")
    assert parse_delta("6m 1x") == timedelta(seconds=360), parse_delta("6m 1x")


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



"""

"""
import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.utils.dates import date_range
import datetime
from rich import traceback

traceback.install()


def test_date_range():
    start = datetime.datetime(2000, 1, 1)
    end = datetime.datetime(2000, 12, 31)
    drange = list(date_range(start, end))

    # it's not all of the last day, 2000 is a leap year so is 365 (366 - 1)
    assert len(drange) == 365 * 24 + 1, len(drange)


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



from ....errors import MissingDependencyError


def zstd(stream):
    """
    Read zstandard compressed files
    """
    import zstandard  # type:ignore

    with zstandard.open(stream, "rb") as file:  # type:ignore
        yield from file.read().split(b"\n")[:-1]


def lzma(stream):
    """
    Read LZMA compressed files
    """
    # lzma should always be present
    import lzma

    with lzma.open(stream, "rb") as file:  # type:ignore
        yield from file


def unzip(stream):
    """
    Read ZIP compressed files
    """
    # zipfile should always be present
    import io
    import zipfile

    from .parallel_reader import KNOWN_EXTENSIONS

    with zipfile.ZipFile(stream, "r") as zip:
        for file_name in zipfile.ZipFile.namelist(zip):
            file = zip.read(file_name)
            # get the extention of the file(s) in the ZIP and put them
            # through a secondary decompressor and parser
            ext = "." + file_name.split(".")[-1]
            if ext in KNOWN_EXTENSIONS:
                decompressor, parser, file_type = KNOWN_EXTENSIONS[ext]
                for line in decompressor(io.BytesIO(file)):
                    yield parser(line)


def parquet(stream):
    """
    Read parquet formatted files
    """
    try:
        import pyarrow.parquet as pq  # type:ignore
    except ImportError:  # pragma: no cover
        raise MissingDependencyError(
            "`pyarrow` is missing, please install or include in requirements.txt"
        )

    table = pq.read_table(stream)
    yield from table.to_pylist()


def lines(stream):
    """
    Default reader, assumes text format
    """
    text = stream.read()  # type:ignore
    yield from text.splitlines()


def block(stream):
    yield stream.read()


def csv(stream):
    import csv

    yield from csv.DictReader(stream.read().decode("utf8").splitlines())



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data import STORAGE_CLASS
from mabel.adapters.disk import DiskReader
from mabel.data.readers.internals.sql_reader import SqlReader
from rich import traceback

traceback.install()


def test_where():
    s = SqlReader(
        "SELECT * FROM tests.data.tweets WHERE username == 'BBCNews'",
        inner_reader=DiskReader,
        raw_path=True,
        persistence=STORAGE_CLASS.MEMORY,
    )
    assert s.count() == 6, s.count()


def test_sql_returned_rows():
    """ """
    # fmt:off
    SQL_TESTS = [
        {"statement":"SELECT * FROM tests.data.index.is", "result":65499},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE user_name = 'Verizon Support'", "result":2},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE user_name == 'Verizon Support'", "result":2},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE `user_name` = 'Verizon Support'", "result":2},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE user_name = \"Verizon Support\"", "result":2},
        {"statement":"select * from tests.data.index.is  where user_name = 'Verizon Support'", "result":2},
        {"statement":"SELECT * FROM tests.data.index.not WHERE user_name = 'Verizon Support'", "result":2},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE user_name = '********'", "result":0},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE user_name LIKE '_erizon _upport'", "result":2},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE user_name LIKE '%Support%'", "result":31},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE user_name = 'Verizon Support'", "result":2},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE tweet_id = 1346604539923853313", "result":1}, 
        {"statement":"SELECT * FROM tests.data.index.is  WHERE tweet_id = 1346604539923853313 AND user_id = 4832862820", "result":1},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE tweet_id IN (1346604539923853313, 1346604544134885378)", "result":2},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE tweet_id = 1346604539923853313 OR user_id = 2147860407", "result":2},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE tweet_id = 1346604539923853313 OR user_verified = True", "result":453},
        {"statement":"SELECT * FROM tests.data.index.is  WHERE user_name = 'Dave Jamieson' AND user_verified = True", "result":1},
        {"statement":"SELECT COUNT(*) FROM tests.data.index.is  WHERE user_name = 'Dave Jamieson' AND user_verified = True", "result":1},
        {"statement":"SELECT count(*) FROM tests.data.index.is GROUP BY user_verified", "result":2},
        {"statement":"SELECT COUNT (*) FROM tests.data.index.is GROUP BY user_verified", "result":2},
        {"statement":"SELECT Count(*) FROM tests.data.index.is GROUP BY user_verified", "result":2},
        {"statement":"SELECT COUNT(*), user_verified FROM tests.data.index.is GROUP BY user_verified", "result":2},
        {"statement":"SELECT * FROM tests.data.index.is WHERE hash_tags contains 'Georgia'", "result":50},
        {"statement":"SELECT COUNT(*) FROM (SELECT user_name FROM tests.data.index.is GROUP BY user_name)", "result":1},
        {"statement":"SELECT MAX(user_name) FROM tests.data.index.is", "result":1},
        {"statement":"SELECT AVG(followers) FROM tests.data.index.is", "result":1},
        {"statement":"SELECT * FROM tests.data.index.is ORDER BY user_name", "result":10000}, # ORDER BY is 10000 record limited
        {"statement":"SELECT * FROM tests.data.index.is ORDER BY user_name ASC", "result":10000},  # ORDER BY is 10000 record limited
        {"statement":"SELECT * FROM tests.data.index.is ORDER BY user_name DESC", "result":10000},  # ORDER BY is 10000 record limited
        {"statement":"SELECT COUNT(user_id) FROM tests.data.index.is", "result":1},
        {"statement":"SELECT * FROM tests.data.index.is WHERE user_id > 1000000", "result":65475},
        {"statement":"SELECT * FROM tests.data.index.is WHERE followers > 100.0", "result":49601},
        {"statement":"SELECT COUNT(*), user_verified, user_id FROM tests.data.index.is GROUP BY user_verified, user_id", "result":60724},
        {"statement":"SELECT * FROM tests.data.index.is WHERE user_name IN ('Steve Strong', 'noel')", "result":3},
        {"statement":"SELECT * FROM tests.data.index.is WHERE (user_name IN ('Steve Strong') OR user_name IN ('noel'))", "result":3},
        {"statement":"SELECT `followers` FROM tests.data.index.is", "result":65499},
        {"statement":"SELECT `user name` FROM tests.data.gaps", "result":25},
        {"statement":"SELECT `user name` FROM tests.data.gaps WHERE `user name` = 'NBCNews'", "result":21},
        {"statement":"SELECT `user_verified` FROM tests.data.gaps", "result":25},
    ]
    # fmt:on

    for test in SQL_TESTS:
        try:
            s = SqlReader(
                test.get("statement"),
                inner_reader=DiskReader,
                raw_path=True,
                persistence=STORAGE_CLASS.MEMORY,
            )
            len_s = len(s.collect_list())
            assert len_s == test.get(
                "result"
            ), f"{test.get('statement')} == {len_s} ({test.get('result')})"
        except Exception as e:
            print(test.get("statement"))
            raise e


def test_sql_to_dictset():
    s = SqlReader(
        sql_statement="SELECT * FROM tests.data.index.not",
        inner_reader=DiskReader,
        raw_path=True,
        persistence=STORAGE_CLASS.MEMORY,
    )
    keys = s.keys()
    assert "tweet_id" in keys, keys
    assert "text" in keys, keys
    assert "followers" in keys, keys
    assert len(s.take(10).collect_list()) == 10


def test_sql_returned_cols():
    # fmt:off
    SQL_TESTS = [
        {"statement":"SELECT tweet_id, user_name FROM tests.data.index.not LIMIT 1", "keys":["tweet_id","user_name"]},
        {"statement":"SELECT COUNT(*) AS count FROM tests.data.index.not LIMIT 1", "keys": ["count"]},
        {"statement":"SELECT COUNT(*) FROM tests.data.index.not LIMIT 1", "keys": ["COUNT(*)"]}
                
    ]
    # fmt:on

    for test in SQL_TESTS:
        s = SqlReader(
            test.get("statement"),
            inner_reader=DiskReader,
            raw_path=True,
            persistence=STORAGE_CLASS.MEMORY,
        )
        # print(s.collect())
        cols = s.keys()
        assert sorted(cols) == sorted(test["keys"]), f"{test.get('statement')} - {cols}"


def test_limit():
    s = SqlReader(
        sql_statement="SELECT tweet_id, user_name FROM tests.data.index.not LIMIT 12",
        inner_reader=DiskReader,
        raw_path=True,
        persistence=STORAGE_CLASS.MEMORY,
    )
    record_count = s.count()
    assert record_count == 12, record_count


def test_group_by_count():
    s = SqlReader(
        sql_statement="SELECT COUNT(*) FROM tests.data.index.not GROUP BY user_verified",
        inner_reader=DiskReader,
        raw_path=True,
    )
    records = s.collect_list()
    record_count = len(records)
    assert record_count == 2, record_count

    s = SqlReader(
        sql_statement="SELECT COUNT(*), user_name FROM tests.data.index.not GROUP BY user_name",
        inner_reader=DiskReader,
        raw_path=True,
    )
    records = s.collect_list()
    record_count = len(records)
    assert record_count == 56527, record_count


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))
from mabel.data import DatabaseWriter
from mabel.data import Reader
from rich import traceback
from mabel.adapters.database import NullWriter

traceback.install()


def finalizer(writer):
    print(f"finalizing {writer}")


def test_database_writer_simple_passed_finalizer():
    w = DatabaseWriter(
        dataset="_temp",
        schema=[{"name": "character", "type": "VARCHAR"}, {"name": "persona", "type": "VARCHAR"}],
        inner_writer=NullWriter,
        engine={},
        finalizer=finalizer,
    )
    for i in range(int(1e5)):
        w.append({"character": "Barney Stinson", "persona": "Lorenzo Von Matterhorn"})
        w.append({"character": "Laszlo Cravensworth", "persona": "Jackie Daytona"})
    number_of_rows = w.finalize()

    assert (
        number_of_rows == 2e5
    ), f"unique database writer expected {2e5} records, found {number_of_rows}"


def test_database_writer_simple_default_finalizer():
    w = DatabaseWriter(
        dataset="_temp",
        schema=[{"name": "character", "type": "VARCHAR"}, {"name": "persona", "type": "VARCHAR"}],
        inner_writer=NullWriter,
        engine={},
    )
    for i in range(int(1e5)):
        w.append({"character": "Barney Stinson", "persona": "Lorenzo Von Matterhorn"})
        w.append({"character": "Laszlo Cravensworth", "persona": "Jackie Daytona"})
    number_of_rows = w.finalize()

    assert (
        number_of_rows == 2e5
    ), f"unique database writer expected {2e5} records, found {number_of_rows}"


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import datetime
import itertools
import re
import threading
import time

from orso.logging import get_logger

from mabel.data.writers.internals.writer_pool import WriterPool
from mabel.data.writers.writer import Writer
from mabel.utils import dates
from mabel.utils import paths
from mabel.utils import text

# shared lock to we're not adding and removing streams at the same time
lock = threading.Lock()


class StreamWriter(Writer):
    """
    Extend the functionality of the Writer to better support streaming data
    """

    def __init__(
        self,
        *,
        dataset: str,
        format: str = "zstd",
        idle_timeout_seconds: int = 30,
        writer_pool_capacity: int = 10,
        **kwargs,
    ):
        """
        Create a Data Writer to write data records into partitions.

        Parameters:
            dataset: string (optional)
                The name of the dataset - this is used to map to a path
            schema: mabel.validator.Schema (optional)
                Schema used to test records for conformity, default is no
                schema and therefore no validation
            format: string (optional)
                - jsonl: raw json lines
                - lzma: lzma compressed json lines
                - zstd: zstandard compressed json lines (default)
                - parquet: Apache Parquet
            idle_timeout_seconds: integer (optional)
                The number of seconds to wait before evicting writers from the
                pool for inactivity, default is 30 seconds
            writer_pool_capacity: integer (optional)
                The number of writers to leave in the writers pool before
                writers are evicted for over capacity, default is 5
            blob_size: integer (optional)
                The maximum size of blobs, the default is 32Mb
            inner_writer: BaseWriter (optional)
                The component used to commit data, the default writer is the
                NullWriter

        Note:
            Different inner_writers may take or require additional parameters.
        """
        if kwargs.get("date"):
            get_logger().warning("Cannot specify a `date` for the StreamWriter.")

        # add the values to kwargs
        kwargs["format"] = format
        kwargs["dataset"] = dataset
        self.dataset = dataset

        super().__init__(**kwargs)

        self.date = dates.parse_iso(kwargs.get("date"))

        self.idle_timeout_seconds = idle_timeout_seconds

        # we have a pool of writers of size maximum_writers
        self.writer_pool_capacity = writer_pool_capacity
        self.writer_pool = WriterPool(pool_size=writer_pool_capacity, **kwargs)

        # establish the background thread responsible for the pool
        self.thread = threading.Thread(target=self.pool_attendant)
        self.thread.name = "mabel-writer-pool-attendant"
        self.thread.daemon = True
        self.run_pool_attendant = True
        self.thread.start()
        get_logger().debug("Pool attendant on-duty")

    def append(self, record: dict):
        """
        Append a new record to the Writer

        Parameters:
            record: dictionary
                The record to append to the Writer

        Returns:
            integer
                The number of records in the current blob
        """
        # get the appropritate writer from the pool and append the record
        # the writer identity is the base of the path where the partitions
        # are written.

        # Check the new record conforms to the schema
        # unlike the batch writer, we don't want to bail out if we have a
        # problem here, instead we're going to save the file to a BACKOUT
        # partition

        writes = 0
        identity = paths.date_format(
            self.dataset_template, (self.date or datetime.datetime.utcnow().date())
        )

        if hasattr(record, "dict"):
            record = record.dict()
        if hasattr(record, "dump_model"):
            record = record.dump_model()

        elif self.schema:
            try:
                self.schema.validate(record)
            except Exception as e:
                identity += "/BACKOUT/"
                get_logger().warning(
                    f"Schema Validation Failed ({e}) - message being written to {identity}"
                )

        with lock:
            # get the placeholders from the dataset name
            placeholders = set(re.findall(r"\{(.*?)\}", identity))

            # there's no substitutions needed, so just write the record
            if len(placeholders) == 0:
                blob_writer = self.writer_pool.get_writer(identity)
                return blob_writer.append(record)

            # get the values from the record, there can be multiple of these
            values = []
            for placeholder in placeholders:
                value = record.get(placeholder)
                if not isinstance(value, list):
                    value = [value]
                values.append(value)
            # get the cartesian product of these lists
            # save the result to a set otherwise it's not a cartesian product
            value_combinations = {i for i in itertools.product(*values)}

            # for every variation in the cartesian product
            for values in value_combinations:  # type:ignore
                this_identity = identity
                # do the actual replacing of the placeholders
                for k, v in zip(placeholders, values):
                    this_identity = this_identity.replace("{" + k + "}", text.sanitize(str(v)))

                # get the writer and save the record
                blob_writer = self.writer_pool.get_writer(this_identity)
                blob_writer.append(record)
                writes += 1

        return writes

    def finalize(self, **kwargs):
        self.run_pool_attendant = False
        with lock:
            for blob_writer_identity in self.writer_pool.writers:
                try:
                    get_logger().debug(
                        f"Removing from the writer pool during finalization, identity={blob_writer_identity['identity']}, poolsize={len(self.writer_pool.writers)}"
                    )
                    self.writer_pool.remove_writer(blob_writer_identity["identity"])
                except Exception as err:
                    get_logger().debug(
                        f"Error finalizing `{blob_writer_identity}`, {type(err).__name__} - {err}"
                    )
        return super().finalize()

    def pool_attendant(self):
        """
        Writer Pool Management
        """
        while self.run_pool_attendant:
            with lock:
                # search for pool occupants who haven't had a write recently
                for blob_writer_identity in self.writer_pool.get_stale_writers(
                    self.idle_timeout_seconds
                ):
                    get_logger().debug(
                        f"Evicting {blob_writer_identity} from the writer pool due to inactivity - limit is {self.idle_timeout_seconds} seconds, poolsize={len(self.writer_pool.writers)}"
                    )
                    self.writer_pool.remove_writer(blob_writer_identity)
                # if we're over capacity, evict the LRU writers
                for blob_writer_identity in self.writer_pool.nominate_writers_to_evict():
                    get_logger().debug(
                        f"Evicting {blob_writer_identity} from the writer pool due the pool being over its {self.writer_pool_capacity} capacity, poolsize={len(self.writer_pool.writers)}"
                    )
                    self.writer_pool.remove_writer(blob_writer_identity)

            time.sleep(0.1)

        get_logger().debug("Pool attendant off-duty")



from ...data.readers.internals.base_inner_reader import BUFFER_SIZE
from ...data.readers.internals.base_inner_reader import BaseInnerReader


class DiskReader(BaseInnerReader):
    def __init__(self, **kwargs):
        """
        File System Reader
        """
        super().__init__(**kwargs)

    def get_blobs_at_path(self, path):
        if path.exists():
            blobs = path.rglob("*")
            return [blob.as_posix() for blob in blobs if blob.is_file()]
        return []

    def get_blob_bytes(self, blob_name: str) -> bytes:
        with open(blob_name, "rb") as f:
            return f.read()



import sys
import os

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from mabel.data.internals.dictset import STORAGE_CLASS, DictSet
from rich import traceback

traceback.install()


def summarize(ds):
    res = []
    for row in ds:
        res.append(row["value"])
    return {"count": len(res), "min": min(res), "max": max(res)}


def test_group_by_advanced():
    data = [
        {"user": "bob", "value": 1},
        {"user": "bob", "value": 2},
        {"user": "bob", "value": 1},
        {"user": "bob", "value": 2},
        {"user": "bob", "value": 1},
        {"user": "alice", "value": 3},
        {"user": "alice", "value": 4},
        {"user": "alice", "value": 3},
        {"user": "alice", "value": 4},
        {"user": "alice", "value": 5},
        {"user": "alice", "value": 5},
        {"user": "eve", "value": 6},
        {"user": "eve", "value": 7},
    ]
    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)

    # don't deduplicate
    groups = ds.collect_set("user", dedupe=False).apply(summarize)
    assert groups["bob"] == {"count": 5, "min": 1, "max": 2}, groups["bob"]
    assert groups["alice"] == {"count": 6, "min": 3, "max": 5}
    assert groups["eve"] == {"count": 2, "min": 6, "max": 7}

    # deduplicate
    groups = ds.collect_set("user", dedupe=True).apply(summarize)
    assert groups["bob"] == {"count": 2, "min": 1, "max": 2}
    assert groups["alice"] == {"count": 3, "min": 3, "max": 5}
    assert groups["eve"] == {"count": 2, "min": 6, "max": 7}


def test_group_by():
    data = [
        {"user": "bob", "value": 1},
        {"user": "bob", "value": 2},
        {"user": "alice", "value": 3},
        {"user": "alice", "value": 4},
        {"user": "alice", "value": 5},
        {"user": "eve", "value": 6},
        {"user": "eve", "value": 7},
    ]

    ds = DictSet(data, storage_class=STORAGE_CLASS.MEMORY)
    groups = ds.collect_set("user")

    # the right number of groups
    assert len(groups) == 3

    # the groups have the right number of records
    assert groups.count("bob") == 2
    assert groups.count("alice") == 3
    assert groups.count("eve") == 2

    # the aggregations work
    assert groups.aggregate("value", max).get("bob") == 2
    assert groups.aggregate("value", min).get("alice") == 3
    assert groups.aggregate("value", sum).get("eve") == 13


if __name__ == "__main__":  # pragma: no cover
    from tests.helpers.runner import run_tests

    run_tests()



import os

from orso.logging.create_logger import get_logger
from urllib3.exceptions import ProtocolError  # type:ignore

from mabel.data.writers.internals.base_inner_writer import BaseInnerWriter
from mabel.errors import MissingDependencyError

try:
    from google.api_core import retry  # type:ignore
    from google.api_core.exceptions import InternalServerError  # type:ignore
    from google.api_core.exceptions import TooManyRequests
    from google.auth.credentials import AnonymousCredentials  # type:ignore
    from google.cloud import storage  # type:ignore

    google_cloud_storage_installed = True
except ImportError:  # pragma: no cover
    google_cloud_storage_installed = False


class GoogleCloudStorageWriter(BaseInnerWriter):
    def __init__(self, credentials=None, **kwargs):
        if not google_cloud_storage_installed:  # pragma: no cover
            raise MissingDependencyError(
                "`google-cloud-storage` is missing, please install or include in requirements.txt"
            )

        super().__init__(**kwargs)
        self.credentials = credentials

        predicate = retry.if_exception_type(
            ConnectionResetError, ProtocolError, InternalServerError, TooManyRequests
        )
        self.retry = retry.Retry(predicate)

    def commit(self, byte_data, override_blob_name=None):
        # this means we're testing
        if os.environ.get("STORAGE_EMULATOR_HOST") is not None:
            client = storage.Client(credentials=AnonymousCredentials())
        else:  # pragma: no cover
            client = storage.Client(credentials=self.credentials)
        self.gcs_bucket = client.get_bucket(self.bucket)
        self.filename = self.filename_without_bucket

        # if we've been given the filename, use that, otherwise get the
        # name from the path builder
        if override_blob_name:
            blob_name = override_blob_name
        else:
            blob_name = self._build_path()

        try:
            blob = self.gcs_bucket.blob(blob_name)
            self.retry(blob.upload_from_string)(byte_data, content_type="application/octet-stream")
            return blob_name
        except Exception as err:  # pragma: no cover
            import traceback

            logger = get_logger()
            logger.error(
                f"Error Saving Blob to GCS {type(err).__name__} - {err}\n{traceback.format_exc()}"
            )
            raise err
